# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-04-13 13:26
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0034_remove_start_time'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='ipmapping',
            name='project',
        ),
        migrations.DeleteModel(
            name='IpMapping',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-06 12:17
from __future__ import unicode_literals

import django.utils.timezone
import model_utils.fields
from django.db import migrations, models

import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('title', models.CharField(max_length=255)),
                ('icon', models.ImageField(blank=True, null=True, upload_to='marketplace_category_icons')),
                ('description', models.TextField(blank=True)),
            ],
            options={
                'verbose_name': 'Category',
                'verbose_name_plural': 'Categories',
                'ordering': ('title',)
            },
        ),
        migrations.CreateModel(
            name='Offering',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('thumbnail', models.ImageField(blank=True, null=True, upload_to='marketplace_service_offering_thumbnails')),
                ('full_description', models.TextField(blank=True)),
                ('rating', models.IntegerField(default=0)),
                ('features', waldur_core.core.fields.JSONField(default=[])),
                ('geolocations', waldur_core.core.fields.JSONField(blank=True, default=[], help_text='List of latitudes and longitudes. For example: [{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]')),
                ('is_active', models.BooleanField(default=True)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offerings', to='marketplace.Category')),
                ('provider', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offerings', to='marketplace.ServiceProvider')),
            ],
            options={
                'verbose_name': 'Offering',
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-27 15:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0068_componentusage_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='componentusage',
            name='uuid',
            field=models.UUIDField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-05-21 10:48
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0031_unique_internal_ip'),
    ]

    operations = [
        migrations.AlterField(
            model_name='internalip',
            name='instance',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='internal_ips_set', to='openstack_tenant.Instance'),
        ),
    ]



from django import forms
from django.contrib import admin
from django.core import urlresolvers

from . import models


class JupyterHubManagementAdminForm(forms.ModelForm):
    class Meta:
        model = models.JupyterHubManagement
        fields = ('session_time_to_live_hours',)
        readonly_fields = ('uuid', 'created', 'python_management', 'user', 'instance_content_type', 'instance_object_id', 'project')


class JupyterHubManagementAdmin(admin.ModelAdmin):
    list_filter = ('session_time_to_live_hours',)
    list_display = ('uuid', 'created', 'instance_content_type', 'instance_object_id', 'python_management_link', 'project', 'user')
    list_display_links = ('uuid',)

    def python_management_link(self, obj):
        link = urlresolvers.reverse("admin:python_management_pythonmanagement_change", args=[obj.id])
        return '<a href="%s">%s</a>' % (link, obj.python_management.virtual_envs_dir_path)

    python_management_link.allow_tags = True
    python_management_link.short_description = 'Python Management'


class JupyterHubOAuthConfigAdminForm(forms.ModelForm):
    class Meta:
        model = models.JupyterHubOAuthConfig
        fields = ('type', 'oauth_callback_url', 'client_id', 'client_secret', 'tenant_id', 'gitlab_host')
        readonly_fields = ('uuid', 'created')


class JupyterHubOAuthConfigAdmin(admin.ModelAdmin):
    list_filter = ('type', 'oauth_callback_url', 'gitlab_host')
    list_display = ('uuid', 'type', 'oauth_callback_url', 'gitlab_host', 'jupyter_hub_management_link')
    list_display_links = ('uuid',)

    def jupyter_hub_management_link(self, obj):
        link = urlresolvers.reverse("admin:jupyter_hub_management_jupyterhubmanagement_change", args=[obj.id])
        return '<a href="%s">JupyterHub management</a>' % link

    jupyter_hub_management_link.allow_tags = True
    jupyter_hub_management_link.short_description = 'JupyterHub Management'


class RequestAdminForm(forms.ModelForm):
    class Meta:
        fields = ('state', 'output')
        readonly_fields = ('uuid', 'created', 'output')


class RequestAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'created', 'state', 'jupyter_hub_management_link')
    list_display_links = ('uuid',)

    def jupyter_hub_management_link(self, obj):
        link = urlresolvers.reverse("admin:jupyter_hub_management_jupyterhubmanagement_change", args=[obj.id])
        return '<a href="%s">JupyterHub management</a>' % link

    jupyter_hub_management_link.allow_tags = True
    jupyter_hub_management_link.short_description = 'JupyterHub Management'


class JupyterHubManagementSyncConfigurationRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.JupyterHubManagementSyncConfigurationRequest


class JupyterHubManagementSyncConfigurationRequestAdmin(RequestAdmin):
    pass


class JupyterHubManagementMakeVirtualEnvironmentGlobalRequestForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest


class JupyterHubManagementMakeVirtualEnvironmentGlobalRequestAdmin(RequestAdmin):
    pass


class JupyterHubManagementDeleteRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.JupyterHubManagementDeleteRequest


class JupyterHubManagementDeleteRequestAdmin(RequestAdmin):
    pass


class JupyterHubManagementMakeVirtualEnvironmentLocalRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest


class JupyterHubManagementMakeVirtualEnvironmentLocalRequestAdmin(RequestAdmin):
    pass


admin.site.register(models.JupyterHubManagement, JupyterHubManagementAdmin)
admin.site.register(models.JupyterHubOAuthConfig, JupyterHubOAuthConfigAdmin)
admin.site.register(models.JupyterHubManagementSyncConfigurationRequest, JupyterHubManagementSyncConfigurationRequestAdmin)
admin.site.register(models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest, JupyterHubManagementMakeVirtualEnvironmentGlobalRequestAdmin)
admin.site.register(models.JupyterHubManagementDeleteRequest, JupyterHubManagementDeleteRequestAdmin)
admin.site.register(models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest, JupyterHubManagementMakeVirtualEnvironmentLocalRequestAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-17 19:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0036_api_secret_code'),
    ]

    operations = [
        migrations.CreateModel(
            name='ComponentUsage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('usage', models.PositiveIntegerField(default=0)),
                ('date', models.DateField()),
                ('component', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.PlanComponent')),
                ('order_item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='usages', to='marketplace.OrderItem')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='componentusage',
            unique_together=set([('order_item', 'component', 'date')]),
        ),
    ]



""" Custom loggers that allows to store logs in DB """

from __future__ import unicode_literals

from collections import defaultdict
import datetime
import decimal
import importlib
import logging
import types
import uuid

from django.apps import apps
import six

from waldur_core.logging import models
from waldur_core.logging.log import EventLoggerAdapter
from waldur_core.logging.middleware import get_event_context

logger = logging.getLogger(__name__)


class LoggerError(AttributeError):
    pass


class EventLoggerError(AttributeError):
    pass


class BaseLogger(object):
    def __init__(self, logger_name=__name__):
        self._meta = getattr(self, 'Meta', None)
        self.supported_types = self.get_supported_types()

    def get_supported_types(self):
        raise NotImplementedError

    def get_supported_groups(self):
        raise NotImplementedError

    def get_nullable_fields(self):
        return getattr(self._meta, 'nullable_fields', [])

    def get_field_model(self, model):
        if not isinstance(model, six.string_types):
            return model

        try:
            return apps.get_model(model)
        except LookupError:
            try:
                app_name, class_name = model.split('.')
                module = importlib.import_module('waldur_core.%s.models' % app_name)
                return getattr(module, class_name)
            except (ImportError, AttributeError, IndexError):
                raise LoggerError("Can't find model %s" % model)

    def compile_message(self, message_template, context):
        try:
            msg = six.text_type(message_template).format(**context)
        except KeyError as e:
            raise LoggerError(
                "Cannot find %s context field. Choices are: %s" % (
                    six.text_type(e), ', '.join(context.keys())))
        return msg

    def validate_logging_type(self, logging_type):
        if self.supported_types and logging_type not in self.supported_types:
            raise EventLoggerError(
                "Unsupported logging type '%s'. Choices are: %s" % (
                    logging_type, ', '.join(self.supported_types)))

    def compile_context(self, **kwargs):
        # Get a list of fields here in order to be sure all models already loaded.
        if not hasattr(self, 'fields'):
            self.fields = {
                k: self.get_field_model(v)
                for k, v in self.__class__.__dict__.items()
                if (not k.startswith('_') and
                    not isinstance(v, types.FunctionType) and
                    not isinstance(v, staticmethod) and k != 'Meta')}

        missed = set(self.fields.keys()) - set(self.get_nullable_fields()) - set(kwargs.keys())
        if missed:
            raise LoggerError("Missed fields in event context: %s" % ', '.join(missed))

        context = {}

        event_context = get_event_context()
        if event_context:
            context.update(event_context)
            username = event_context.get('user_username')
            if 'user' in self.fields and username:
                logger.warning("User is passed directly to event context. "
                               "Currently authenticated user %s is ignored.", username)

        for entity_name, entity in six.iteritems(kwargs):
            if entity_name in self.fields:
                entity_class = self.fields[entity_name]
                if entity is None and entity_name in self.get_nullable_fields():
                    continue
                if not isinstance(entity, entity_class):
                    raise LoggerError(
                        "Field '%s' must be an instance of %s but %s received" % (
                            entity_name, entity_class.__name__, entity.__class__.__name__))
            else:
                logger.error(
                    "Field '%s' cannot be used in logging context for %s",
                    entity_name, self.__class__.__name__)
                continue

            if isinstance(entity, LoggableMixin):
                context.update(entity._get_log_context(entity_name))
            elif isinstance(entity, (int, float, six.string_types, dict, tuple, list, bool)):
                context[entity_name] = entity
            elif entity is None:
                pass
            else:
                context[entity_name] = six.text_type(entity)
                logger.warning(
                    "Cannot properly serialize '%s' context field. "
                    "Must be inherited from LoggableMixin." % entity_name)

        return context


class EventLogger(BaseLogger):
    """ Base event logger API.
        Fields which must be passed during event log emitting (event context)
        should be defined as attributes for this class in the form of:

        field_name = ObjectClass || '<app_label>.<class_name>'

        A list of supported event types can be defined with help of method get_supported_types,
        or 'event_types' property of Meta class. Event type won't be validated if this list is empty.

        Example usage:

        .. code-block:: python

            from waldur_core.openstack.models import Tenant

            class QuotaEventLogger(EventLogger):
                tenant = Tenant
                project = 'structure.Project'
                threshold = float
                quota_type = six.string_types

                class Meta:
                    event_types = 'quota_threshold_reached',


            quota_logger = QuotaEventLogger(__name__)
            quota_logger.warning(
                '{quota_type} quota threshold has been reached for {project_name}.',
                event_type='quota_threshold_reached',
                event_context=dict(
                    quota_type=quota.name,
                    project=membership.project,
                    tenant=tenant,
                    threshold=threshold * quota.limit)
            )
    """

    def __init__(self, logger_name=__name__):
        super(EventLogger, self).__init__(logger_name)
        self.logger = EventLoggerAdapter(logging.getLogger(logger_name))

    def get_supported_types(self):
        return getattr(self._meta, 'event_types', tuple())

    def get_supported_groups(self):
        return getattr(self._meta, 'event_groups', {})

    @staticmethod
    def get_scopes(event_context):
        """
        This method receives event context and returns set of Django model objects.
        For example, if resource is specified in event context then it should return
        resource, project and customer.
        """
        return set()

    def info(self, *args, **kwargs):
        self.process('info', *args, **kwargs)

    def error(self, *args, **kwargs):
        self.process('error', *args, **kwargs)

    def warning(self, *args, **kwargs):
        self.process('warning', *args, **kwargs)

    def debug(self, *args, **kwargs):
        self.process('debug', *args, **kwargs)

    def process(self, level, message_template, event_type='undefined', event_context=None):
        self.validate_logging_type(event_type)

        if not event_context:
            event_context = {}

        context = self.compile_context(**event_context)
        msg = self.compile_message(message_template, context)
        log = getattr(self.logger, level)
        log(msg, extra={'event_type': event_type, 'event_context': context})

        event = models.Event.objects.create(
            event_type=event_type,
            message=msg,
            context=context,
        )
        if event_context:
            for scope in self.get_scopes(event_context):
                if scope and scope.id:
                    models.Feed.objects.create(scope=scope, event=event)


class LoggableMixin(object):
    """ Mixin to serialize model in logs.
        Extends django model or custom class with fields extraction method.
    """

    def get_log_fields(self):
        return ('uuid', 'name')

    def _get_log_context(self, entity_name=None):

        context = {}
        for field in self.get_log_fields():
            if not hasattr(self, field):
                continue

            value = getattr(self, field)

            if entity_name:
                name = "{}_{}".format(entity_name, field)
            else:
                name = field
            if isinstance(value, uuid.UUID):
                context[name] = value.hex
            elif isinstance(value, LoggableMixin):
                context.update(value._get_log_context(field))
            elif isinstance(value, datetime.date):
                context[name] = value.isoformat()
            elif isinstance(value, decimal.Decimal):
                context[name] = float(value)
            elif isinstance(value, dict):
                context[name] = value
            elif callable(value):
                context[name] = value()
            else:
                context[name] = six.text_type(value)

        return context

    @classmethod
    def get_permitted_objects(cls, user):
        return cls.objects.none()


class BaseLoggerRegistry(object):

    def get_loggers(self):
        raise NotImplementedError('Method "get_loggers" is not implemented.')

    def register(self, name, logger_class):
        if name in self.__dict__:
            raise EventLoggerError("Logger '%s' already registered." % name)
        self.__dict__[name] = logger_class()

    def unregister_all(self):
        self.__dict__ = {}

    def get_all_types(self):
        events = set()
        for elogger in self.get_loggers():
            events.update(elogger.get_supported_types())
        return list(sorted(events))

    def get_all_groups(self):
        groups = defaultdict(set)
        for elogger in self.get_loggers():
            for group, items in elogger.get_supported_groups().items():
                groups[group].update(set(items))
        return dict(groups)

    def expand_groups(self, groups):
        items = set()
        mapping = self.get_all_groups()
        for group in groups:
            items.update(mapping.get(group, []))
        return sorted(items)


class EventLoggerRegistry(BaseLoggerRegistry):

    def get_loggers(self):
        return [l for l in self.__dict__.values() if isinstance(l, EventLogger)]


def get_valid_events():
    return event_logger.get_all_types()


def get_event_groups():
    return event_logger.get_all_groups()


def get_event_groups_keys():
    return sorted(get_event_groups().keys())


def expand_event_groups(groups):
    return event_logger.expand_groups(groups)


class CustomEventLogger(EventLogger):
    """
    Logger for custom events e.g. created by staff user
    """
    scope = LoggableMixin

    class Meta:
        event_types = ('custom_notification',)
        nullable_fields = ('scope',)


# This global objects represent the default loggers registry
event_logger = EventLoggerRegistry()

event_logger.register('custom', CustomEventLogger)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-05 13:57
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0026_remove_type_protect_template'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offeringtemplate',
            name='name',
            field=models.CharField(max_length=150, verbose_name='name'),
        ),
    ]



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import permissions, viewsets

from waldur_core.structure import filters as structure_filters

from . import filters, models, serializers


class SlurmPackageViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.SlurmPackage.objects.all()
    serializer_class = serializers.SlurmPackageSerializer
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend,)
    filter_class = filters.SlurmPackageFilter



from __future__ import unicode_literals

from waldur_core.structure import views as structure_views

from . import filters, executors, models, serializers


class ServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.RijkscloudService.objects.all()
    serializer_class = serializers.ServiceSerializer


class ServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.RijkscloudServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filter_class = filters.ServiceProjectLinkFilter


class FlavorViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Flavor.objects.all().order_by('cores', 'ram')
    serializer_class = serializers.FlavorSerializer
    lookup_field = 'uuid'
    filter_class = filters.FlavorFilter


class VolumeViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Volume.objects.all()
    serializer_class = serializers.VolumeSerializer
    filter_class = filters.VolumeFilter
    create_executor = executors.VolumeCreateExecutor
    pull_executor = executors.VolumePullExecutor
    delete_executor = executors.VolumeDeleteExecutor
    disabled_actions = ['update', 'partial_update']

    importable_resources_backend_method = 'get_volumes_for_import'
    importable_resources_serializer_class = serializers.VolumeImportableSerializer
    import_resource_serializer_class = serializers.VolumeImportSerializer


class InstanceViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Instance.objects.all()
    serializer_class = serializers.InstanceSerializer
    filter_class = filters.InstanceFilter
    pull_executor = executors.InstancePullExecutor
    create_executor = executors.InstanceCreateExecutor
    delete_executor = executors.InstanceDeleteExecutor
    disabled_actions = ['update', 'partial_update']

    importable_resources_backend_method = 'get_instances_for_import'
    importable_resources_serializer_class = serializers.InstanceImportableSerializer
    import_resource_serializer_class = serializers.InstanceImportSerializer


class NetworkViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Network.objects.all().order_by('settings')
    serializer_class = serializers.NetworkSerializer
    lookup_field = 'uuid'
    filter_class = filters.NetworkFilter


class SubNetViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.SubNet.objects.all().order_by('settings')
    serializer_class = serializers.SubNetSerializer
    lookup_field = 'uuid'
    filter_class = filters.SubNetFilter


class InternalIPViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.InternalIP.objects.all().order_by('settings', 'address')
    serializer_class = serializers.InternalIPSerializer
    lookup_field = 'uuid'
    filter_class = filters.InternalIPFilter


class FloatingIPViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.FloatingIP.objects.all().order_by('settings', 'address')
    serializer_class = serializers.FloatingIPSerializer
    lookup_field = 'uuid'
    filter_class = filters.FloatingIPFilter



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-26 17:12
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0004_issue_types'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='issue',
            name='impact',
        ),
        migrations.RemoveField(
            model_name='project',
            name='impact_field',
        ),
    ]



from . import views


def register_in(router):
    router.register(r'rijkscloud', views.ServiceViewSet, base_name='rijkscloud')
    router.register(r'rijkscloud-service-project-link', views.ServiceProjectLinkViewSet,
                    base_name='rijkscloud-spl')
    router.register(r'rijkscloud-flavors', views.FlavorViewSet, base_name='rijkscloud-flavor')
    router.register(r'rijkscloud-volumes', views.VolumeViewSet, base_name='rijkscloud-volume')
    router.register(r'rijkscloud-instances', views.InstanceViewSet, base_name='rijkscloud-instance')
    router.register(r'rijkscloud-networks', views.NetworkViewSet, base_name='rijkscloud-network')
    router.register(r'rijkscloud-subnets', views.SubNetViewSet, base_name='rijkscloud-subnet')
    router.register(r'rijkscloud-internal-ips', views.InternalIPViewSet, base_name='rijkscloud-internal-ip')
    router.register(r'rijkscloud-floating-ips', views.FloatingIPViewSet, base_name='rijkscloud-fip')



from __future__ import unicode_literals

import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def migrate_data(apps, schema_editor):
    OfferingItem = apps.get_model('invoices', 'OfferingItem')

    class States(object):
        REQUESTED = 'requested'
        OK = 'ok'
        TERMINATED = 'terminated'

    OfferingItem.objects.filter(offering__state=States.REQUESTED).delete()


class Migration(migrations.Migration):
    dependencies = [
        ('invoices', '0020_migrate_payment_details_data'),
    ]

    operations = [
        migrations.RunPython(migrate_data, reverse_code=migrations.RunPython.noop),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-15 13:37
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0064_remove_resource_count_project_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='resource',
            name='cost',
            field=models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True),
        ),
        migrations.AlterField(
            model_name='orderitem',
            name='limits',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-04-12 06:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0040_unique_floating_ip'),
    ]

    operations = [
        migrations.AddField(
            model_name='tenant',
            name='default_volume_type_name',
            field=models.CharField(blank=True, help_text='Volume type name to use when creating volumes.', max_length=100),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-05 12:22
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0002_immutable_default_json'),
        ('marketplace', '0040_categorycolumn'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectResourceCount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('count', models.PositiveIntegerField(default=0)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Category')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='projectresourcecount',
            unique_together=set([('project', 'category')]),
        ),
    ]



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class MarketplaceOpenStackExtension(WaldurExtension):

    class Settings:
        WALDUR_MARKETPLACE_OPENSTACK = {
            'TENANT_CATEGORY_UUID': None,
            'INSTANCE_CATEGORY_UUID': None,
            'VOLUME_CATEGORY_UUID': None,
            'AUTOMATICALLY_CREATE_PRIVATE_OFFERING': True,
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_openstack'

    @staticmethod
    def get_public_settings():
        return [
            'TENANT_CATEGORY_UUID',
            'INSTANCE_CATEGORY_UUID',
            'VOLUME_CATEGORY_UUID',
        ]

    @staticmethod
    def is_assembly():
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-14 11:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0034_plan_quota'),
    ]

    operations = [
        migrations.AddField(
            model_name='plancomponent',
            name='description',
            field=models.CharField(blank=True, max_length=500, verbose_name='description'),
        ),
        migrations.AddField(
            model_name='plancomponent',
            name='measured_unit',
            field=models.CharField(default='', help_text='Unit of measurement, for example, GB.', max_length=30, blank=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='plancomponent',
            name='name',
            field=models.CharField(default='', help_text='Display name for the measured unit, for example, Floating IP.', max_length=150),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='plancomponent',
            name='type',
            field=models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50),
        ),
    ]



from __future__ import unicode_literals

from django.http import Http404, HttpResponse
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status, exceptions
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from waldur_core.core import views as core_views
from waldur_core.core import utils as core_utils
from waldur_core.structure import filters as structure_filters, permissions as structure_permissions
from . import filters, models, serializers, tasks


class InvoiceViewSet(core_views.ReadOnlyActionsViewSet):
    queryset = models.Invoice.objects.order_by('-year', '-month')
    serializer_class = serializers.InvoiceSerializer
    lookup_field = 'uuid'
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.InvoiceFilter

    def _is_invoice_created(invoice):
        if invoice.state != models.Invoice.States.CREATED:
            raise exceptions.ValidationError(_('Notification only for the created invoice can be sent.'))

    @detail_route(methods=['post'])
    def send_notification(self, request, uuid=None):
        invoice = self.get_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        link_template = serializer.validated_data.get('link_template')
        tasks.send_invoice_notification.delay(invoice.uuid.hex, link_template)
        tasks.create_invoice_pdf.delay(core_utils.serialize_instance(invoice))

        return Response({'detail': _('Invoice notification sending has been successfully scheduled.')},
                        status=status.HTTP_200_OK)

    send_notification_serializer_class = serializers.InvoiceNotificationSerializer
    send_notification_permissions = [structure_permissions.is_staff]
    send_notification_validators = [_is_invoice_created]

    @detail_route()
    def pdf(self, request, uuid=None):
        invoice = self.get_object()
        if not invoice.has_file():
            tasks.create_invoice_pdf.delay(core_utils.serialize_instance(invoice))
            raise Http404()

        file_response = HttpResponse(invoice.file, content_type='application/pdf')
        filename = invoice.get_filename()
        file_response['Content-Disposition'] = 'attachment; filename="{filename}"'.format(filename=filename)
        return file_response



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-22 10:01
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_enlarge_username'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='details',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Extra details from authentication backend.'),
        ),
    ]



from __future__ import unicode_literals

import logging
import re

import six
from django.core import validators as django_validators
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers, models as structure_models, SupportedServices

from . import models, executors
from .backend import JiraBackendError

logger = logging.getLogger(__name__)


class ServiceSerializer(structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': 'JIRA host (e.g. https://jira.example.com/)',
        'username': 'JIRA user with excessive privileges',
        'password': '',
        'token': '',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.JiraService
        view_name = 'jira-detail'


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.JiraServiceProjectLink
        view_name = 'jira-spl-detail'
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'jira-detail'},
        }


class BaseJiraPropertySerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = NotImplemented
        fields = ('url', 'uuid', 'name', 'description', 'icon_url')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class ProjectTemplateSerializer(BaseJiraPropertySerializer):

    class Meta(BaseJiraPropertySerializer.Meta):
        model = models.ProjectTemplate


class IssueTypeSerializer(BaseJiraPropertySerializer):

    class Meta(BaseJiraPropertySerializer.Meta):
        model = models.IssueType
        fields = BaseJiraPropertySerializer.Meta.fields + ('subtask',)


class PrioritySerializer(BaseJiraPropertySerializer):

    class Meta(BaseJiraPropertySerializer.Meta):
        model = models.Priority


class ProjectSerializer(structure_serializers.BaseResourceSerializer):
    key = serializers.CharField(write_only=True, validators=[
        django_validators.RegexValidator(
            regex=re.compile('[A-Z][A-Z0-9]+'),
            message=_('Project keys must start with an uppercase letter, '
                      'followed by one or more uppercase alphanumeric characters.'),
        ),
        django_validators.MaxLengthValidator(
            limit_value=10,
            message=_('The project key must not exceed 10 characters in length.')
        ),
    ])

    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='jira-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='jira-spl-detail',
        queryset=models.JiraServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    template = serializers.HyperlinkedRelatedField(
        view_name='jira-project-templates-detail',
        queryset=models.ProjectTemplate.objects.all(),
        lookup_field='uuid'
    )

    template_name = serializers.ReadOnlyField(source='template.name')
    template_description = serializers.ReadOnlyField(source='template.description')
    issue_types = IssueTypeSerializer(many=True, read_only=True)
    priorities = PrioritySerializer(many=True, read_only=True)
    percentage = serializers.SerializerMethodField()

    def get_percentage(self, prj):
        if prj.state not in (models.Project.States.OK,
                             models.Project.States.ERRED):
            return prj.action_details.get('percentage', 0)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Project
        view_name = 'jira-projects-detail'
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'key', 'template',
        )
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'key', 'template', 'template_name', 'template_description',
            'issue_types', 'priorities', 'percentage',
        )

    def create(self, validated_data):
        validated_data['backend_id'] = validated_data['key']
        return super(ProjectSerializer, self).create(validated_data)


class ProjectImportableSerializer(core_serializers.AugmentedSerializerMixin,
                                  serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='jira-spl-detail',
        queryset=models.JiraServiceProjectLink.objects.all(),
        write_only=True)

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta(object):
        model = models.Project
        model_fields = ('name',)
        fields = ('service_project_link', 'backend_id') + model_fields
        read_only_fields = model_fields


class ProjectImportSerializer(ProjectImportableSerializer):
    class Meta(ProjectImportableSerializer.Meta):
        fields = ProjectImportableSerializer.Meta.fields + ('url', 'uuid', 'created',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Project.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({
                'backend_id': _('Project has been imported already.')
            })

        try:
            backend = service_project_link.get_backend()
            project = backend.import_project_scheduled(backend_id, service_project_link)
        except JiraBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import project with ID %s") % validated_data['backend_id']
            })

        executors.ProjectPullExecutor.execute(project)
        return project


class JiraPropertySerializer(core_serializers.RestrictedSerializerMixin,
                             core_serializers.AugmentedSerializerMixin,
                             serializers.HyperlinkedModelSerializer):
    state = serializers.ReadOnlyField(source='get_state_display')

    class Meta(object):
        model = NotImplemented
        fields = (
            'url', 'uuid', 'user', 'user_uuid', 'user_name', 'user_email', 'state', 'error_message', 'backend_id'
        )
        read_only_fields = 'uuid', 'user', 'error_message', 'backend_id'
        extra_kwargs = {
            'user': {'lookup_field': 'uuid', 'view_name': 'user-detail'},
        }
        related_paths = {
            'user': ('uuid', 'name', 'email'),
        }

    def create(self, validated_data):
        validated_data['user'] = self.context['request'].user
        return super(JiraPropertySerializer, self).create(validated_data)


class CommentSerializer(JiraPropertySerializer):

    class Meta(JiraPropertySerializer.Meta):
        model = models.Comment
        fields = JiraPropertySerializer.Meta.fields + (
            'issue', 'issue_uuid', 'issue_key', 'message', 'created',
        )
        protected_fields = 'issue',
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'jira-comments-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'jira-issues-detail'},
            **JiraPropertySerializer.Meta.extra_kwargs
        )
        related_paths = dict(
            issue=('uuid', 'key'),
            **JiraPropertySerializer.Meta.related_paths
        )


class AttachmentSerializer(JiraPropertySerializer):

    class Meta(JiraPropertySerializer.Meta):
        model = models.Attachment
        fields = JiraPropertySerializer.Meta.fields + (
            'issue', 'issue_uuid', 'issue_key', 'file', 'created',
        )
        protected_fields = 'issue',
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'jira-attachments-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'jira-issues-detail'},
            **JiraPropertySerializer.Meta.extra_kwargs
        )
        related_paths = dict(
            issue=('uuid', 'key'),
            **JiraPropertySerializer.Meta.related_paths
        )


class IssueSerializer(JiraPropertySerializer):
    priority = serializers.HyperlinkedRelatedField(
        view_name='jira-priorities-detail',
        queryset=models.Priority.objects.all(),
        lookup_field='uuid',
    )
    access_url = serializers.ReadOnlyField(source='get_access_url')
    comments = CommentSerializer(many=True, read_only=True)

    scope = core_serializers.GenericRelatedField(
        source='resource',
        related_models=structure_models.ResourceMixin.get_all_models(),
        required=False
    )
    scope_type = serializers.SerializerMethodField()
    scope_name = serializers.ReadOnlyField(source='resource.name')

    parent = serializers.HyperlinkedRelatedField(
        view_name='jira-issues-detail',
        queryset=models.Issue.objects.all(),
        lookup_field='uuid',
        required=False,
        allow_null=True,
    )

    # For consistency with resource serializer render
    # Waldur project as project and JIRA project as jira_project
    project = serializers.HyperlinkedRelatedField(
        source='project.service_project_link.project',
        view_name='project-detail',
        read_only=True,
        lookup_field='uuid'
    )

    project_name = serializers.ReadOnlyField(source='project.service_project_link.project.name')
    project_uuid = serializers.ReadOnlyField(source='project.service_project_link.project.uuid')

    jira_project = serializers.HyperlinkedRelatedField(
        queryset=models.Project.objects.all(),
        source='project',
        view_name='jira-projects-detail',
        lookup_field='uuid'
    )

    jira_project_name = serializers.ReadOnlyField(source='project.name')
    jira_project_uuid = serializers.ReadOnlyField(source='project.uuid')

    resource_type = serializers.SerializerMethodField()
    service_settings_state = serializers.SerializerMethodField()

    def get_resource_type(self, obj):
        return 'JIRA.Issue'

    def get_service_settings_state(self, obj):
        return 'OK'

    def get_scope_type(self, obj):
        if obj.resource:
            return SupportedServices.get_name_for_model(obj.resource_content_type.model_class())

    class Meta(JiraPropertySerializer.Meta):
        model = models.Issue
        fields = JiraPropertySerializer.Meta.fields + (
            'project', 'project_uuid', 'project_name',
            'jira_project', 'jira_project_uuid', 'jira_project_name',
            'key', 'summary', 'description', 'resolution', 'status',
            'priority', 'priority_name', 'priority_icon_url', 'priority_description',
            'created', 'updated',
            'creator_username', 'creator_name', 'creator_email',
            'assignee_username', 'assignee_name', 'assignee_email',
            'reporter_username', 'reporter_name', 'reporter_email',
            'resolution_date',
            'access_url', 'comments', 'resource_type', 'service_settings_state',
            'type', 'type_name', 'type_description', 'type_icon_url',
            'scope', 'scope_type', 'scope_name',
            'parent', 'parent_uuid', 'parent_summary', 'resolution_sla',
        )
        read_only_fields = 'status', 'resolution', 'updated_username', 'error_message', 'resolution_sla', 'backend_id'
        protected_fields = 'jira_project', 'key', 'type', 'scope',
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'jira-issues-detail'},
            type={'lookup_field': 'uuid', 'view_name': 'jira-issue-types-detail'},
            parent={'lookup_field': 'uuid', 'view_name': 'jira-issues-detail'},
            **JiraPropertySerializer.Meta.extra_kwargs
        )
        related_paths = dict(
            type=('icon_url', 'name', 'description'),
            parent=('uuid', 'summary'),
            priority=('icon_url', 'name', 'description'),
            **JiraPropertySerializer.Meta.related_paths
        )

    def create(self, validated_data):
        project = validated_data['project']
        issue_type = validated_data['type']
        if issue_type not in project.issue_types.all():
            valid_choices = ', '.join(project.issue_types.values_list('name', flat=True))
            raise serializers.ValidationError({
                'type': _('Invalid issue type. Please select one of following: %s') % valid_choices
            })

        priority = validated_data['priority']
        if priority.settings != project.service_project_link.service.settings:
            raise serializers.ValidationError({
                'parent': _('Priority should belong to the same JIRA provider.')
            })

        parent_issue = validated_data.get('parent')
        if parent_issue:
            if not issue_type.subtask:
                raise serializers.ValidationError({
                    'parent': _('Issue type is not subtask, parent issue is not allowed.')
                })

            if parent_issue.project != project:
                raise serializers.ValidationError({
                    'parent': _('Parent issue should belong to the same JIRA project.')
                })

        return super(IssueSerializer, self).create(validated_data)


#
# Serializers below are used by webhook only
#

class JiraCommentSerializer(serializers.Serializer):
    id = serializers.CharField()


class JiraChangelogSerializer(serializers.Serializer):
    items = serializers.ListField()


class JiraFieldSerializer(serializers.Serializer):
    id = serializers.CharField()
    name = serializers.CharField()


class JiraIssueProjectSerializer(JiraFieldSerializer):
    key = serializers.CharField()


class JiraIssueFieldsSerializer(serializers.Serializer):
    project = JiraIssueProjectSerializer()
    comment = serializers.DictField(required=False)


class JiraIssueSerializer(serializers.Serializer):
    key = serializers.CharField()
    fields = JiraIssueFieldsSerializer()


class WebHookReceiverSerializer(serializers.Serializer):
    class Event:
        ISSUE_CREATE = 1
        ISSUE_UPDATE = 2
        ISSUE_DELETE = 4
        COMMENT_CREATE = 5
        COMMENT_UPDATE = 6
        COMMENT_DELETE = 7

        ISSUE_ACTIONS = (ISSUE_CREATE, ISSUE_UPDATE, ISSUE_DELETE)
        COMMENT_ACTIONS = (COMMENT_CREATE, COMMENT_UPDATE, COMMENT_DELETE)

        CHOICES = {
            ('jira:issue_created', ISSUE_CREATE),
            ('jira:issue_updated', ISSUE_UPDATE),
            ('jira:issue_deleted', ISSUE_DELETE),
            ('comment_created', COMMENT_CREATE),
            ('comment_updated', COMMENT_UPDATE),
            ('comment_deleted', COMMENT_DELETE),
        }

    @classmethod
    def remove_event(cls, events):
        if isinstance(events, six.text_type):
            events = [events]

        elements = set(filter(lambda e: e[0] in events, cls.Event.CHOICES))
        cls.Event.CHOICES.difference_update(elements)

    webhookEvent = serializers.ChoiceField(choices=Event.CHOICES)
    issue = JiraIssueSerializer()
    comment = JiraCommentSerializer(required=False)
    changelog = JiraChangelogSerializer(required=False)
    issue_event_type_name = serializers.CharField(required=False)  # For old Jira's version

    def get_project(self, project_key):
        try:
            project = models.Project.objects.get(backend_id=project_key)
        except models.Project.DoesNotExist:
            raise serializers.ValidationError('Project with id %s does not exist.' % project_key)
        return project

    def get_issue(self, project, key, create):
        issue = None

        try:
            issue = models.Issue.objects.get(project=project, backend_id=key)
        except models.Issue.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Issue with id %s does not exist.' % key)

        return issue

    def get_comment(self, issue, key, create):
        comment = None

        try:
            comment = models.Comment.objects.get(issue=issue, backend_id=key)
        except models.Comment.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Comment with id %s does not exist.' % key)

        return comment

    def create(self, validated_data):
        event_type = dict(self.Event.CHOICES).get(validated_data['webhookEvent'])
        fields = validated_data['issue']['fields']
        key = validated_data['issue']['key']
        project_key = fields['project']['key']
        project = self.get_project(project_key)
        backend = project.get_backend()
        create_issue = event_type == self.Event.ISSUE_CREATE
        issue = self.get_issue(project, key, create_issue)

        if fields.get('comment', False):
            # The processing of hooks requests for the old and new Jira versions is different.
            # The main difference is that in the old version, when changing comments,
            # jira:issue_updated event is sent to the new comment_X event.
            old_jira = validated_data.get('issue_event_type_name', True)
        else:
            old_jira = False

        if event_type in self.Event.ISSUE_ACTIONS:
            if not issue and create_issue:
                backend.create_issue_from_jira(project, key)

            if event_type == self.Event.ISSUE_UPDATE:
                if old_jira:
                    if old_jira == 'issue_commented':
                        comment_backend_id = validated_data['comment']['id']
                        backend.create_comment_from_jira(issue, comment_backend_id)

                    if old_jira == 'issue_comment_edited':
                        comment_backend_id = validated_data['comment']['id']
                        comment = self.get_comment(issue, comment_backend_id, False)
                        backend.update_comment_from_jira(comment)

                    if old_jira == 'issue_comment_deleted':
                        backend.delete_old_comments(issue)

                    if old_jira in ('issue_updated', 'issue_generic'):
                        new_attachment = filter(lambda x: x['field'] == 'Attachment',
                                                validated_data['changelog']['items'])
                        if new_attachment:
                            backend.update_attachment_from_jira(issue)

                        backend.update_issue_from_jira(issue)

                else:
                    new_attachment = filter(lambda x: x['fieldId'] == 'attachment',
                                            validated_data['changelog']['items'])

                    if new_attachment:
                        backend.update_attachment_from_jira(issue)

                    backend.update_issue_from_jira(issue)

            if event_type == self.Event.ISSUE_DELETE:
                backend.delete_issue_from_jira(issue)

        if event_type in self.Event.COMMENT_ACTIONS:
            try:
                comment_backend_id = validated_data['comment']['id']
            except KeyError:
                raise serializers.ValidationError('Request not include fields.comment.id')

            create_comment = event_type == self.Event.COMMENT_CREATE
            comment = self.get_comment(issue, comment_backend_id, create_comment)

            if not comment and create_comment:
                backend.create_comment_from_jira(issue, comment_backend_id)

            if event_type == self.Event.COMMENT_UPDATE:
                backend.update_comment_from_jira(comment)

            if event_type == self.Event.COMMENT_DELETE:
                backend.delete_comment_from_jira(comment)

        return validated_data



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-08 08:07
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('support', '0027_non_unique_template_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='RequestBasedOffering',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('support.offering',),
        ),
    ]



from __future__ import unicode_literals

from django.contrib.auth.decorators import login_required

from waldur_core.core import views

from .log import event_logger


@login_required
def login_completed(request):
    """
    Callback view called after user has successfully logged in.
    Redirects user agent to frontend view with valid token as hash parameter.
    """
    token = views.RefreshTokenMixin().refresh_token(request.user)
    event_logger.auth_openid.info(
        'User {user_full_name} authenticated successfully with eID.',
        event_type='auth_logged_in_with_openid',
        event_context={'user': request.user}
    )
    return views.login_completed(token.key, 'openid')


def login_failed(request, message):
    return views.login_failed(message)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-04 09:02
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import django_fsm
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0011_order_item'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offering',
            name='attributes',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict,
                                                                 help_text='Fields describing Category.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-03-22 14:51
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0054_payment_details'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('python_management', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonmanagement',
            name='instance_content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='pythonmanagement',
            name='instance_object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='pythonmanagement',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='pythonmanagement',
            name='system_user',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='virtualenvironment',
            name='jupyter_hub_global',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterUniqueTogether(
            name='pythonmanagement',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='pythonmanagement',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='pythonmanagement',
            name='service_project_link',
        ),
    ]



from __future__ import unicode_literals

import logging
import time
from collections import defaultdict
from functools import partial

from django.conf import settings as django_settings
from django.contrib import auth
from django.db import transaction, IntegrityError
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters as rf_filters
from rest_framework import mixins, views, viewsets, status
from rest_framework import permissions as rf_permissions
from rest_framework import serializers as rf_serializers
from rest_framework.decorators import detail_route, list_route
from rest_framework.exceptions import PermissionDenied, MethodNotAllowed, NotFound, APIException, ValidationError
from rest_framework.response import Response
from reversion.models import Version
import six

from waldur_core.core import managers as core_managers
from waldur_core.core import mixins as core_mixins
from waldur_core.core import models as core_models
from waldur_core.core import serializers as core_serializers
from waldur_core.core import signals as core_signals
from waldur_core.core import validators as core_validators
from waldur_core.core import views as core_views
from waldur_core.core.utils import datetime_to_timestamp, sort_dict
from waldur_core.logging import models as logging_models
from waldur_core.quotas.models import QuotaModelMixin, Quota
from waldur_core.structure import (
    SupportedServices, ServiceBackendError, ServiceBackendNotImplemented,
    filters, managers, models, permissions, serializers, utils)
from waldur_core.structure.managers import filter_queryset_for_user
from waldur_core.structure.metadata import ActionsMetadata
from waldur_core.structure.signals import resource_imported, structure_role_updated

logger = logging.getLogger(__name__)

User = auth.get_user_model()


class CustomerViewSet(core_mixins.EagerLoadMixin, viewsets.ModelViewSet):
    queryset = models.Customer.objects.all().order_by('name')
    serializer_class = serializers.CustomerSerializer
    lookup_field = 'uuid'
    filter_backends = (filters.GenericUserFilter,
                       filters.GenericRoleFilter,
                       DjangoFilterBackend,
                       rf_filters.OrderingFilter,
                       filters.AccountingStartDateFilter,
                       filters.ExternalCustomerFilterBackend,)
    ordering_fields = (
        'abbreviation',
        'accounting_start_date',
        'contact_details',
        'created',
        'name',
        'native_name',
        'registration_code',
    )
    filter_class = filters.CustomerFilter

    def list(self, request, *args, **kwargs):
        """
        To get a list of customers, run GET against */api/customers/* as authenticated user. Note that a user can
        only see connected customers:

        - customers that the user owns
        - customers that have a project where user has a role

        Staff also can filter customers by user UUID, for example /api/customers/?user_uuid=<UUID>

        Staff also can filter customers by exists accounting_start_date, for example:

        The first category:
        /api/customers/?accounting_is_running=True
            has accounting_start_date empty (i.e. accounting starts at once)
            has accounting_start_date in the past (i.e. has already started).

        Those that are not in the first:
        /api/customers/?accounting_is_running=False # exists accounting_start_date

        """
        return super(CustomerViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        Optional `field` query parameter (can be list) allows to limit what fields are returned.
        For example, given request /api/customers/<uuid>/?field=uuid&field=name you get response like this:

        .. code-block:: javascript

            {
                "uuid": "90bcfe38b0124c9bbdadd617b5d739f5",
                "name": "Ministry of Bells"
            }
        """
        return super(CustomerViewSet, self).retrieve(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        A new customer can only be created:

         - by users with staff privilege (is_staff=True);
         - by any user if OWNER_CAN_MANAGE_CUSTOMER is set to True;

        If user who has created new organization is not staff, he is granted owner permission.

        Example of a valid request:

        .. code-block:: http

            POST /api/customers/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "Customer A",
                "native_name": "Customer A",
                "abbreviation": "CA",
                "contact_details": "Luhamaa 28, 10128 Tallinn",
            }
        """
        return super(CustomerViewSet, self).create(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """
        Deletion of a customer is done through sending a **DELETE** request to the customer instance URI. Please note,
        that if a customer has connected projects, deletion request will fail with 409 response code.

        Valid request example (token is user specific):

        .. code-block:: http

            DELETE /api/customers/6c9b01c251c24174a6691a1f894fae31/ HTTP/1.1
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com
        """
        return super(CustomerViewSet, self).destroy(request, *args, **kwargs)

    def get_serializer_class(self):
        if self.action == 'users':
            return serializers.CustomerUserSerializer
        return super(CustomerViewSet, self).get_serializer_class()

    def get_serializer_context(self):
        context = super(CustomerViewSet, self).get_serializer_context()
        if self.action == 'users':
            context['customer'] = self.get_object()
        return context

    def check_customer_permissions(self, customer=None):
        if self.request.user.is_staff:
            return

        if not django_settings.WALDUR_CORE.get('OWNER_CAN_MANAGE_CUSTOMER'):
            raise PermissionDenied()

        if not customer:
            return

        if not customer.has_user(self.request.user, models.CustomerRole.OWNER):
            raise PermissionDenied()

    def perform_create(self, serializer):
        self.check_customer_permissions()
        customer = serializer.save()
        if not self.request.user.is_staff:
            customer.add_user(self.request.user, models.CustomerRole.OWNER, self.request.user)

        if django_settings.WALDUR_CORE.get('CREATE_DEFAULT_PROJECT_ON_ORGANIZATION_CREATION', False):
            project = models.Project(
                name=_('First project'),
                description=_('First project we have created for you'),
                customer=customer,
            )
            project.save()

    def perform_update(self, serializer):
        self.check_customer_permissions(serializer.instance)
        utils.check_customer_blocked(serializer.instance)
        return super(CustomerViewSet, self).perform_update(serializer)

    def perform_destroy(self, instance):
        self.check_customer_permissions(instance)
        utils.check_customer_blocked(instance)

        core_signals.pre_delete_validate.send(
            sender=models.Customer,
            instance=instance,
            user=self.request.user
        )

        return super(CustomerViewSet, self).perform_destroy(instance)

    @detail_route(filter_backends=[filters.GenericRoleFilter])
    def users(self, request, uuid=None):
        """ A list of users connected to the customer. """
        customer = self.get_object()
        queryset = customer.get_users()
        # we need to handle filtration manually because we want to filter only customer users, not customers.
        filter_backend = filters.UserConcatenatedNameOrderingBackend()
        queryset = filter_backend.filter_queryset(request, queryset, self)
        queryset = self.paginate_queryset(queryset)
        serializer = self.get_serializer(queryset, many=True)
        return self.get_paginated_response(serializer.data)


class ProjectTypeViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.ProjectType.objects.all()
    serializer_class = serializers.ProjectTypeSerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.ProjectTypeFilter


class ProjectViewSet(core_mixins.EagerLoadMixin, core_views.ActionsViewSet):
    queryset = models.Project.objects.all().order_by('name')
    serializer_class = serializers.ProjectSerializer
    lookup_field = 'uuid'
    filter_backends = (
        filters.GenericRoleFilter,
        DjangoFilterBackend,
        filters.CustomerAccountingStartDateFilter,
    )
    filter_class = filters.ProjectFilter
    destroy_validators = partial_update_validators = [utils.check_customer_blocked]

    def get_serializer_context(self):
        context = super(ProjectViewSet, self).get_serializer_context()
        if self.action == 'users':
            context['project'] = self.get_object()
        return context

    def list(self, request, *args, **kwargs):
        """
        To get a list of projects, run **GET** against */api/projects/* as authenticated user.
        Here you can also check actual value for project quotas and project usage

        Note that a user can only see connected projects:

        - projects that the user owns as a customer
        - projects where user has any role

        Supported logic filters:

        - ?can_manage - return a list of projects where current user is manager or a customer owner;
        - ?can_admin - return a list of projects where current user is admin;
        """
        return super(ProjectViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        Optional `field` query parameter (can be list) allows to limit what fields are returned.
        For example, given request /api/projects/<uuid>/?field=uuid&field=name you get response like this:

        .. code-block:: javascript

            {
                "uuid": "90bcfe38b0124c9bbdadd617b5d739f5",
                "name": "Default"
            }
        """
        return super(ProjectViewSet, self).retrieve(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        A new project can be created by users with staff privilege (is_staff=True) or customer owners.
        Project resource quota is optional. Example of a valid request:

        .. code-block:: http

            POST /api/projects/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "Project A",
                "customer": "http://example.com/api/customers/6c9b01c251c24174a6691a1f894fae31/",
            }
        """
        return super(ProjectViewSet, self).create(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """
        Deletion of a project is done through sending a **DELETE** request to the project instance URI.
        Please note, that if a project has connected instances, deletion request will fail with 409 response code.

        Valid request example (token is user specific):

        .. code-block:: http

            DELETE /api/projects/6c9b01c251c24174a6691a1f894fae31/ HTTP/1.1
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com
        """
        return super(ProjectViewSet, self).destroy(request, *args, **kwargs)

    def can_create_project_with(self, customer):
        user = self.request.user

        if user.is_staff:
            return True

        if customer.has_user(user, models.CustomerRole.OWNER):
            return True

        return False

    def get_queryset(self):
        user = self.request.user
        queryset = super(ProjectViewSet, self).get_queryset()

        can_manage = self.request.query_params.get('can_manage', None)
        if can_manage is not None:
            queryset = queryset.filter(
                Q(customer__permissions__user=user,
                  customer__permissions__role=models.CustomerRole.OWNER,
                  customer__permissions__is_active=True) |
                Q(permissions__user=user,
                  permissions__role=models.ProjectRole.MANAGER,
                  permissions__is_active=True)
            ).distinct()

        can_admin = self.request.query_params.get('can_admin', None)

        if can_admin is not None:
            queryset = queryset.filter(
                permissions__user=user,
                permissions__role=models.ProjectRole.ADMINISTRATOR,
                permissions__is_active=True
            )

        return queryset

    def perform_create(self, serializer):
        customer = serializer.validated_data['customer']

        if not self.can_create_project_with(customer):
            raise PermissionDenied()

        utils.check_customer_blocked(customer)

        customer.validate_quota_change({'nc_project_count': 1}, raise_exception=True)

        super(ProjectViewSet, self).perform_create(serializer)

    @detail_route(filter_backends=[filters.GenericRoleFilter])
    def users(self, request, uuid=None):
        """ A list of users connected to the project """
        project = self.get_object()
        queryset = project.get_users()
        # we need to handle filtration manually because we want to filter only project users, not projects.
        filter_backend = filters.UserConcatenatedNameOrderingBackend()
        queryset = filter_backend.filter_queryset(request, queryset, self)
        queryset = self.paginate_queryset(queryset)
        serializer = self.get_serializer(queryset, many=True)
        return self.get_paginated_response(serializer.data)

    users_serializer_class = serializers.ProjectUserSerializer

    @detail_route(methods=['post'])
    def update_certifications(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()
        serialized_instance = serializers.ProjectSerializer(instance, context={'request': self.request})

        return Response(serialized_instance.data, status=status.HTTP_200_OK)

    update_certifications_serializer_class = serializers.ServiceCertificationsUpdateSerializer
    update_certifications_permissions = [permissions.is_owner]


class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = serializers.UserSerializer
    lookup_field = 'uuid'
    permission_classes = (
        rf_permissions.IsAuthenticated,
        permissions.IsAdminOrOwner,
    )
    filter_backends = (
        filters.CustomerUserFilter,
        filters.ProjectUserFilter,
        filters.UserFilterBackend,
        DjangoFilterBackend,
    )
    filter_class = filters.UserFilter

    def list(self, request, *args, **kwargs):
        """
        User list is available to all authenticated users. To get a list,
        issue authenticated **GET** request against */api/users/*.

        User list supports several filters. All filters are set in HTTP query section.
        Field filters are listed below. All of the filters apart from ?organization are
        using case insensitive partial matching.

        Several custom filters are supported:

        - ?current - filters out user making a request. Useful for getting information about a currently logged in user.
        - ?civil_number=XXX - filters out users with a specified civil number
        - ?is_active=True|False - show only active (non-active) users

        The user can be created either through automated process on login with SAML token, or through a REST call by a user
        with staff privilege.

        Example of a creation request is below.

        .. code-block:: http

            POST /api/users/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "username": "sample-user",
                "full_name": "full name",
                "native_name": "taisnimi",
                "job_title": "senior cleaning manager",
                "email": "example@example.com",
                "civil_number": "12121212",
                "phone_number": "",
                "description": "",
                "organization": "",
            }

        NB! Username field is case-insensitive. So "John" and "john" will be treated as the same user.
        """
        return super(UserViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        User fields can be updated by account owner or user with staff privilege (is_staff=True).
        Following user fields can be updated:

        - organization (deprecated, use
          `organization plugin <http://waldur_core-organization.readthedocs.org/en/stable/>`_ instead)
        - full_name
        - native_name
        - job_title
        - phone_number
        - email

        Can be done by **PUT**ing a new data to the user URI, i.e. */api/users/<UUID>/* by staff user or account owner.
        Valid request example (token is user specific):

        .. code-block:: http

            PUT /api/users/e0c058d06864441fb4f1c40dee5dd4fd/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "email": "example@example.com",
                "organization": "Bells organization",
            }
        """
        return super(UserViewSet, self).retrieve(request, *args, **kwargs)

    @detail_route(methods=['post'])
    def password(self, request, uuid=None):
        """
        To change a user password, submit a **POST** request to the user's RPC URL, specifying new password
        by staff user or account owner.

        Password is expected to be at least 7 symbols long and contain at least one number
        and at least one lower or upper case.

        Example of a valid request:

        .. code-block:: http

            POST /api/users/e0c058d06864441fb4f1c40dee5dd4fd/password/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "password": "nQvqHzeP123",
            }
        """
        user = self.get_object()

        serializer = serializers.PasswordSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        new_password = serializer.validated_data['password']
        user.set_password(new_password)
        user.save()

        return Response({'detail': _('Password has been successfully updated.')},
                        status=status.HTTP_200_OK)


class BasePermissionViewSet(viewsets.ModelViewSet):
    """
    This is a base class for both customer and project permissions.
    scope_field is required parameter, it should be either 'customer' or 'project'.
    quota_scope_field is optional parameter, it is used in order to validate quotas on permission creation.
    """

    scope_field = None
    quota_scope_field = None

    def perform_create(self, serializer):
        scope = serializer.validated_data[self.scope_field]
        role = serializer.validated_data['role']
        affected_user = serializer.validated_data['user']
        expiration_time = serializer.validated_data.get('expiration_time')

        if not scope.can_manage_role(self.request.user, role, expiration_time):
            raise PermissionDenied()

        utils.check_customer_blocked(scope)

        if self.quota_scope_field:
            quota_scope = getattr(scope, self.quota_scope_field)
        else:
            quota_scope = scope
        if not quota_scope.get_users().filter(pk=affected_user.pk).exists():
            quota_scope.validate_quota_change({'nc_user_count': 1}, raise_exception=True)

        super(BasePermissionViewSet, self).perform_create(serializer)

    def perform_update(self, serializer):
        permission = serializer.instance
        scope = getattr(permission, self.scope_field)
        role = permission.role

        utils.check_customer_blocked(scope)

        new_expiration_time = serializer.validated_data.get('expiration_time')
        old_expiration_time = permission.expiration_time
        if new_expiration_time == old_expiration_time:
            return

        if not scope.can_manage_role(self.request.user, role, new_expiration_time):
            raise PermissionDenied()

        serializer.save()
        structure_role_updated.send(
            sender=self.queryset.model,
            instance=permission,
            user=self.request.user,
        )

    def perform_destroy(self, instance):
        permission = instance
        scope = getattr(permission, self.scope_field)
        role = permission.role
        affected_user = permission.user
        expiration_time = permission.expiration_time

        if not scope.can_manage_role(self.request.user, role, expiration_time):
            raise PermissionDenied()

        utils.check_customer_blocked(scope)

        scope.remove_user(affected_user, role, removed_by=self.request.user)


class ProjectPermissionViewSet(BasePermissionViewSet):
    """
    - Projects are connected to customers, whereas the project may belong to one customer only,
      and the customer may have
      multiple projects.
    - Projects are connected to services, whereas the project may contain multiple services,
      and the service may belong to multiple projects.
    - Staff members can list all available projects of any customer and create new projects.
    - Customer owners can list all projects that belong to any of the customers they own.
      Customer owners can also create projects for the customers they own.
    - Project administrators can list all the projects they are administrators in.
    - Project managers can list all the projects they are managers in.
    """
    # See CustomerPermissionViewSet for implementation details.

    queryset = models.ProjectPermission.objects.filter(is_active=True).order_by('-created')
    serializer_class = serializers.ProjectPermissionSerializer
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend,)
    filter_class = filters.ProjectPermissionFilter
    scope_field = 'project'
    quota_scope_field = 'customer'

    def list(self, request, *args, **kwargs):
        """
        Project permissions expresses connection of user to a project.
        User may have either project manager or system administrator permission in the project.
        Use */api/project-permissions/* endpoint to maintain project permissions.

        Note that project permissions can be viewed and modified only by customer owners and staff users.

        To list all visible permissions, run a **GET** query against a list.
        Response will contain a list of project users and their brief data.

        To add a new user to the project, **POST** a new relationship to */api/project-permissions/* endpoint specifying
        project, user and the role of the user ('admin' or 'manager'):

        .. code-block:: http

            POST /api/project-permissions/ HTTP/1.1
            Accept: application/json
            Authorization: Token 95a688962bf68678fd4c8cec4d138ddd9493c93b
            Host: example.com

            {
                "project": "http://example.com/api/projects/6c9b01c251c24174a6691a1f894fae31/",
                "role": "manager",
                "user": "http://example.com/api/users/82cec6c8e0484e0ab1429412fe4194b7/"
            }
        """
        return super(ProjectPermissionViewSet, self).list(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """
        To remove a user from a project, delete corresponding connection (**url** field). Successful deletion
        will return status code 204.

        .. code-block:: http

            DELETE /api/project-permissions/42/ HTTP/1.1
            Authorization: Token 95a688962bf68678fd4c8cec4d138ddd9493c93b
            Host: example.com
        """
        return super(ProjectPermissionViewSet, self).destroy(request, *args, **kwargs)


class ProjectPermissionLogViewSet(mixins.RetrieveModelMixin,
                                  mixins.ListModelMixin,
                                  viewsets.GenericViewSet):
    queryset = models.ProjectPermission.objects.filter(is_active=None)
    serializer_class = serializers.ProjectPermissionLogSerializer
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend,)
    filter_class = filters.ProjectPermissionFilter


class CustomerPermissionViewSet(BasePermissionViewSet):
    """
    - Customers are connected to users through roles, whereas user may have role "customer owner".
    - Each customer may have multiple owners, and each user may own multiple customers.
    - Staff members can list all available customers and create new customers.
    - Customer owners can list all customers they own. Customer owners can also create new customers.
    - Project administrators can list all the customers that own any of the projects they are administrators in.
    - Project managers can list all the customers that own any of the projects they are managers in.
    """
    queryset = models.CustomerPermission.objects.filter(is_active=True).order_by('-created')
    serializer_class = serializers.CustomerPermissionSerializer
    filter_class = filters.CustomerPermissionFilter
    scope_field = 'customer'

    def get_queryset(self):
        queryset = super(CustomerPermissionViewSet, self).get_queryset()

        if not (self.request.user.is_staff or self.request.user.is_support):
            queryset = queryset.filter(
                Q(user=self.request.user, is_active=True) |
                Q(customer__projects__permissions__user=self.request.user, is_active=True) |
                Q(customer__permissions__user=self.request.user, is_active=True)
            ).distinct()

        return queryset

    def list(self, request, *args, **kwargs):
        """
        Each customer is associated with a group of users that represent customer owners. The link is maintained
        through **api/customer-permissions/** endpoint.

        To list all visible links, run a **GET** query against a list.
        Response will contain a list of customer owners and their brief data.

        To add a new user to the customer, **POST** a new relationship to **customer-permissions** endpoint:

        .. code-block:: http

            POST /api/customer-permissions/ HTTP/1.1
            Accept: application/json
            Authorization: Token 95a688962bf68678fd4c8cec4d138ddd9493c93b
            Host: example.com

            {
                "customer": "http://example.com/api/customers/6c9b01c251c24174a6691a1f894fae31/",
                "role": "owner",
                "user": "http://example.com/api/users/82cec6c8e0484e0ab1429412fe4194b7/"
            }
        """
        return super(CustomerPermissionViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        To remove a user from a customer owner group, delete corresponding connection (**url** field).
        Successful deletion will return status code 204.

        .. code-block:: http

            DELETE /api/customer-permissions/71/ HTTP/1.1
            Authorization: Token 95a688962bf68678fd4c8cec4d138ddd9493c93b
            Host: example.com
        """
        return super(CustomerPermissionViewSet, self).retrieve(request, *args, **kwargs)


class CustomerPermissionLogViewSet(mixins.RetrieveModelMixin,
                                   mixins.ListModelMixin,
                                   viewsets.GenericViewSet):
    queryset = models.CustomerPermission.objects.filter(is_active=None)
    serializer_class = serializers.CustomerPermissionLogSerializer
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend,)
    filter_class = filters.CustomerPermissionFilter


class CreationTimeStatsView(views.APIView):
    """
    Historical information about creation time of projects and customers.
    """

    def get(self, request, format=None):
        month = 60 * 60 * 24 * 30
        data = {
            'start_timestamp': request.query_params.get('from', int(time.time() - month)),
            'end_timestamp': request.query_params.get('to', int(time.time())),
            'segments_count': request.query_params.get('datapoints', 6),
            'model_name': request.query_params.get('type', 'customer'),
        }

        serializer = serializers.CreationTimeStatsSerializer(data=data)
        serializer.is_valid(raise_exception=True)

        stats = serializer.get_stats(request.user)
        return Response(stats, status=status.HTTP_200_OK)

    def list(self, request, *args, **kwargs):
        """
        Available request parameters:

        - ?type=type_of_statistics_objects (required. Have to be from the list: 'customer', 'project')
        - ?from=timestamp (default: now - 30 days, for example: 1415910025)
        - ?to=timestamp (default: now, for example: 1415912625)
        - ?datapoints=how many data points have to be in answer (default: 6)

        Answer will be list of datapoints(dictionaries).
        Each datapoint will contain fields: 'to', 'from', 'value'.
        'Value' - count of objects, that were created between 'from' and 'to' dates.

        Example:

        .. code-block:: javascript

            [
                {"to": 471970877, "from": 1, "value": 5},
                {"to": 943941753, "from": 471970877, "value": 0},
                {"to": 1415912629, "from": 943941753, "value": 3}
            ]
        """
        return super(CreationTimeStatsView, self).list(request, *args, **kwargs)


class SshKeyViewSet(mixins.CreateModelMixin,
                    mixins.RetrieveModelMixin,
                    mixins.DestroyModelMixin,
                    mixins.ListModelMixin,
                    viewsets.GenericViewSet):
    """
    SSH public keys are injected to VM instances during creation, so that holder of corresponding SSH private key can
    log in to that instance.
    SSH public keys are connected to user accounts, whereas the key may belong to one user only,
    and the user may have multiple SSH keys.
    Users can only access SSH keys connected to their accounts. Staff users can see all the accounts.
    Project administrators can select what SSH key will be injected into VM instance during instance provisioning.
    """

    queryset = core_models.SshPublicKey.objects.all()
    serializer_class = serializers.SshKeySerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.SshKeyFilter

    def get_queryset(self):
        queryset = super(SshKeyViewSet, self).get_queryset()
        if self.request.user.is_staff or self.request.user.is_support:
            return queryset

        return queryset.filter(Q(user=self.request.user) | Q(is_shared=True))

    def perform_destroy(self, instance):
        if instance.is_shared and not self.request.user.is_staff:
            raise PermissionDenied(_('Only staff users are allowed to delete shared SSH public key.'))
        else:
            instance.delete()

    def list(self, request, *args, **kwargs):
        """
        To get a list of SSH keys, run **GET** against */api/keys/* as authenticated user.

        A new SSH key can be created by any active users. Example of a valid request:

        .. code-block:: http

            POST /api/keys/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "ssh_public_key1",
                "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDDURXDP5YhOQUYoDuTxJ84DuzqMJYJqJ8+SZT28
                               TtLm5yBDRLKAERqtlbH2gkrQ3US58gd2r8H9jAmQOydfvgwauxuJUE4eDpaMWupqquMYsYLB5f+vVGhdZbbzfc6DTQ2rY
                               dknWoMoArlG7MvRMA/xQ0ye1muTv+mYMipnd7Z+WH0uVArYI9QBpqC/gpZRRIouQ4VIQIVWGoT6M4Kat5ZBXEa9yP+9du
                               D2C05GX3gumoSAVyAcDHn/xgej9pYRXGha4l+LKkFdGwAoXdV1z79EG1+9ns7wXuqMJFHM2KDpxAizV0GkZcojISvDwuh
                               vEAFdOJcqjyyH4FOGYa8usP1 jhon@example.com",
            }
        """
        return super(SshKeyViewSet, self).list(request, *args, **kwargs)

    def perform_create(self, serializer):
        user = self.request.user
        name = serializer.validated_data['name']

        if core_models.SshPublicKey.objects.filter(user=user, name=name).exists():
            raise rf_serializers.ValidationError({'name': [_('This field must be unique.')]})

        serializer.save(user=user)


class ServiceSettingsViewSet(core_mixins.EagerLoadMixin,
                             core_views.ActionsViewSet):
    queryset = models.ServiceSettings.objects.filter().order_by('pk')
    serializer_class = serializers.ServiceSettingsSerializer
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend,
                       filters.ServiceSettingsScopeFilterBackend,
                       rf_filters.OrderingFilter)
    filter_class = filters.ServiceSettingsFilter
    lookup_field = 'uuid'
    ordering_fields = ('type', 'name', 'state',)
    disabled_actions = ['create', 'destroy']

    def list(self, request, *args, **kwargs):
        """
        To get a list of service settings, run **GET** against */api/service-settings/* as an authenticated user.
        Only settings owned by this user or shared settings will be listed.

        Supported filters are:

        - ?name=<text> - partial matching used for searching
        - ?type=<type> - choices: OpenStack, DigitalOcean, Amazon, JIRA, GitLab, Oracle
        - ?state=<state> - choices: New, Creation Scheduled, Creating, Sync Scheduled, Syncing, In Sync, Erred
        - ?shared=<bool> - allows to filter shared service settings
        """
        return super(ServiceSettingsViewSet, self).list(request, *args, **kwargs)

    def can_user_update_settings(request, view, obj=None):
        """ Only staff can update shared settings, otherwise user has to be an owner of the settings."""
        if obj is None:
            return

        # TODO [TM:3/21/17] clean it up after WAL-634. Clean up service settings update tests as well.
        if obj.customer and not obj.shared:
            return permissions.is_owner(request, view, obj)
        else:
            return permissions.is_staff(request, view, obj)

    def update(self, request, *args, **kwargs):
        """
        To update service settings, issue a **PUT** or **PATCH** to */api/service-settings/<uuid>/* as a customer owner.
        You are allowed to change name and credentials only.

        Example of a request:

        .. code-block:: http

            PATCH /api/service-settings/9079705c17d64e6aa0af2e619b0e0702/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "username": "admin",
                "password": "new_secret"
            }
        """
        return super(ServiceSettingsViewSet, self).update(request, *args, **kwargs)

    update_permissions = partial_update_permissions = [can_user_update_settings,
                                                       permissions.check_access_to_services_management]

    update_validators = partial_update_validators = [utils.check_customer_blocked]

    @detail_route()
    def stats(self, request, uuid=None):
        """
        This endpoint returns allocation of resources for current service setting.
        Answer is service-specific dictionary. Example output for OpenStack:

        * vcpu - maximum number of vCPUs (from hypervisors)
        * vcpu_quota - maximum number of vCPUs(from quotas)
        * vcpu_usage - current number of used vCPUs

        * ram - total size of memory for allocation (from hypervisors)
        * ram_quota - maximum number of memory (from quotas)
        * ram_usage - currently used memory size on all physical hosts

        * storage - total available disk space on all physical hosts (from hypervisors)
        * storage_quota - maximum number of storage (from quotas)
        * storage_usage - currently used storage on all physical hosts

        {
            'vcpu': 10,
            'vcpu_quota': 7,
            'vcpu_usage': 5,
            'ram': 1000,
            'ram_quota': 700,
            'ram_usage': 500,
            'storage': 10000,
            'storage_quota': 7000,
            'storage_usage': 5000
        }
        """

        service_settings = self.get_object()
        backend = service_settings.get_backend()

        try:
            stats = backend.get_stats()
        except ServiceBackendNotImplemented:
            stats = {}

        return Response(stats, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def update_certifications(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()
        serialized_instance = serializers.ServiceSettingsSerializer(instance, context={'request': self.request})

        return Response(serialized_instance.data, status=status.HTTP_200_OK)

    update_certifications_serializer_class = serializers.ServiceCertificationsUpdateSerializer
    update_certifications_permissions = [
        can_user_update_settings,
        permissions.check_access_to_services_management
    ]


class ServiceMetadataViewSet(viewsets.GenericViewSet):
    # Fix for schema generation
    queryset = []

    def list(self, request):
        """
        To get a list of supported service types, run **GET** against */api/service-metadata/* as an authenticated user.
        Use an endpoint from the returned list in order to create new service.
        """
        return Response(SupportedServices.get_services_with_resources(request))


class ResourceSummaryViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    """
    Use */api/resources/* to get a list of all the resources of any type that a user can see.
    """
    model = models.NewResource  # for permissions definition.
    serializer_class = serializers.SummaryResourceSerializer
    filter_backends = (filters.GenericRoleFilter, filters.ResourceSummaryFilterBackend, filters.TagsFilter)

    def get_queryset(self):
        resource_models = {k: v for k, v in SupportedServices.get_resource_models().items()}
        resource_models = self._filter_by_category(resource_models)
        resource_models = self._filter_by_types(resource_models)
        resource_models = self._filter_resources(resource_models)

        queryset = managers.ResourceSummaryQuerySet(resource_models.values())
        return serializers.SummaryResourceSerializer.eager_load(queryset, self.request)

    def _filter_by_types(self, resource_models):
        types = self.request.query_params.getlist('resource_type', None)
        if types:
            resource_models = {k: v for k, v in resource_models.items() if k in types}
        return resource_models

    def _filter_by_category(self, resource_models):
        choices = {
            'apps': models.ApplicationMixin.get_all_models(),
            'vms': models.VirtualMachine.get_all_models(),
            'private_clouds': models.PrivateCloud.get_all_models(),
            'storages': models.Storage.get_all_models(),
            'volumes': models.Volume.get_all_models(),
            'snapshots': models.Snapshot.get_all_models(),
        }
        category = self.request.query_params.get('resource_category')
        if not category:
            return resource_models

        category_models = choices.get(category)
        if category_models:
            return {k: v for k, v in resource_models.items() if v in category_models}
        return {}

    def _filter_resources(self, resource_models):
        return {k: v for k, v in resource_models.items()
                if v in models.ResourceMixin.get_all_models()}

    @transaction.atomic
    def list(self, request, *args, **kwargs):
        """
        To get a list of supported resources' actions, run **OPTIONS** against
        */api/<resource_url>/* as an authenticated user.

        It is possible to filter and order by resource-specific fields, but this filters will be applied only to
        resources that support such filtering. For example it is possible to sort resource by ?o=ram, but SugarCRM crms
        will ignore this ordering, because they do not support such option.

        Filter resources by type or category
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        There are two query argument to select resources by their type.

        - Specify explicitly list of resource types, for example:

          /api/<resource_endpoint>/?resource_type=DigitalOcean.Droplet&resource_type=OpenStack.Instance

        - Specify category, one of vms, apps, private_clouds or storages for example:

          /api/<resource_endpoint>/?category=vms

        Filtering by monitoring fields
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        Resources may have SLA attached to it. Example rendering of SLA:

        .. code-block:: javascript

            "sla": {
                "value": 95.0
                "agreed_value": 99.0,
                "period": "2016-03"
            }

        You may filter or order resources by SLA. Default period is current year and month.

        - Example query for filtering list of resources by actual SLA:

          /api/<resource_endpoint>/?actual_sla=90&period=2016-02

        - Warning! If resource does not have SLA attached to it, it is not included in ordered response.
          Example query for ordering list of resources by actual SLA:

          /api/<resource_endpoint>/?o=actual_sla&period=2016-02

        Service list is displaying current SLAs for each of the items. By default,
        SLA period is set to the current month. To change the period pass it as a query argument:

        - ?period=YYYY-MM - return a list with SLAs for a given month
        - ?period=YYYY - return a list with SLAs for a given year

        In all cases all currently running resources are returned, if SLA for the given period is
        not known or not present, it will be shown as **null** in the response.

        Resources may have monitoring items attached to it. Example rendering of monitoring items:

        .. code-block:: javascript

            "monitoring_items": {
               "application_state": 1
            }

        You may filter or order resources by monitoring item.

        - Example query for filtering list of resources by installation state:

          /api/<resource_endpoint>/?monitoring__installation_state=1

        - Warning! If resource does not have monitoring item attached to it, it is not included in ordered response.
          Example query for ordering list of resources by installation state:

          /api/<resource_endpoint>/?o=monitoring__installation_state

        Filtering by tags
        ^^^^^^^^^^^^^^^^^

        Resource may have tags attached to it. Example of tags rendering:

        .. code-block:: javascript

            "tags": [
                "license-os:centos7",
                "os-family:linux",
                "license-application:postgresql",
                "support:premium"
            ]

        Tags filtering:

         - ?tag=IaaS - filter by full tag name, using method OR. Can be list.
         - ?rtag=os-family:linux - filter by full tag name, using AND method. Can be list.
         - ?tag__license-os=centos7 - filter by tags with particular prefix.

        Tags ordering:

         - ?o=tag__license-os - order by tag with particular prefix. Instances without given tag will not be returned.
        """

        return super(ResourceSummaryViewSet, self).list(request, *args, **kwargs)

    @list_route()
    def count(self, request):
        """
        Count resources by type. Example output:

        .. code-block:: javascript

            {
                "Amazon.Instance": 0,
                "GitLab.Project": 3,
                "Azure.VirtualMachine": 0,
                "DigitalOcean.Droplet": 0,
                "OpenStack.Instance": 0,
                "GitLab.Group": 8
            }
        """
        queryset = self.filter_queryset(self.get_queryset())
        return Response({SupportedServices.get_name_for_model(qs.model): qs.count()
                         for qs in queryset.querysets})


class ServicesViewSet(mixins.ListModelMixin,
                      viewsets.GenericViewSet):
    """ The summary list of all user services. """

    model = models.Service
    serializer_class = serializers.SummaryServiceSerializer
    filter_backends = (filters.GenericRoleFilter, filters.ServiceSummaryFilterBackend)

    def get_queryset(self):
        service_models = {k: v['service'] for k, v in SupportedServices.get_service_models().items()}
        service_models = self._filter_by_types(service_models)
        # TODO: filter models by service type.
        queryset = managers.ServiceSummaryQuerySet(service_models.values())
        return serializers.SummaryServiceSerializer.eager_load(queryset, self.request)

    def _filter_by_types(self, service_models):
        types = self.request.query_params.getlist('service_type', None)
        if types:
            service_models = {k: v for k, v in service_models.items() if k in types}
        return service_models

    def list(self, request, *args, **kwargs):
        """
        Filter services by type
        ^^^^^^^^^^^^^^^^^^^^^^^

        It is possible to filter services by their types. Example:

          /api/services/?service_type=DigitalOcean&service_type=OpenStack
        """
        return super(ServicesViewSet, self).list(request, *args, **kwargs)


class BaseCounterView(viewsets.GenericViewSet):
    # Fix for schema generation
    queryset = []
    extra_counters = {}
    dynamic_counters = set()

    @classmethod
    def register_counter(cls, name, func):
        cls.extra_counters[name] = func

    @classmethod
    def register_dynamic_counter(cls, func):
        cls.dynamic_counters.add(func)

    def get_counters(self):
        counters = self.get_fields()
        for name, func in self.extra_counters.items():
            counters[name] = partial(func, self.object)
        return counters

    def list(self, request, uuid=None):
        result = {}
        counters = self.get_counters()
        for field, func in counters.items():
            result[field] = func()
        for func in self.dynamic_counters:
            result.update(func(self.object))
        fields = request.query_params.getlist('fields')
        if fields:
            result = {k: v for k, v in result.items() if k in fields}
        return Response(result)

    def get_fields(self):
        raise NotImplementedError()

    @cached_property
    def object(self):
        return self.get_object()


class CustomerCountersView(BaseCounterView):
    """
    Count number of entities related to customer

    .. code-block:: javascript

        {
            "services": 1,
            "projects": 1,
            "users": 3
        }
    """
    lookup_field = 'uuid'
    extra_counters = {}
    dynamic_counters = set()

    def get_queryset(self):
        return filter_queryset_for_user(models.Customer.objects.all().only('pk', 'uuid'), self.request.user)

    def get_fields(self):
        return {
            'projects': self.get_projects,
            'services': self.get_services,
            'users': self.get_users
        }

    def get_users(self):
        return self.object.get_users().count()

    def get_projects(self):
        return self._count_model(models.Project)

    def get_services(self):
        models = [item['service'] for item in SupportedServices.get_service_models().values()]
        return self._total_count(models)

    def _total_count(self, models):
        return sum(self._count_model(model) for model in models)

    def _count_model(self, model):
        qs = model.objects.filter(customer=self.object).only('pk')
        qs = filter_queryset_for_user(qs, self.request.user)
        return qs.count()


class ProjectCountersView(BaseCounterView):
    """
    Count number of entities related to project

    .. code-block:: javascript

        {
            "users": 0,
            "apps": 0,
            "vms": 1,
            "private_clouds": 1,
            "storages": 2,
        }
    """
    lookup_field = 'uuid'
    extra_counters = {}
    dynamic_counters = set()

    def get_queryset(self):
        return filter_queryset_for_user(models.Project.objects.all().only('pk', 'uuid'), self.request.user)

    def get_fields(self):
        fields = {
            'vms': self.get_vms,
            'apps': self.get_apps,
            'private_clouds': self.get_private_clouds,
            'storages': self.get_storages,
            'users': self.get_users
        }
        return fields

    def get_vms(self):
        return self._total_count(models.VirtualMachine.get_all_models())

    def get_apps(self):
        return self._total_count(models.ApplicationMixin.get_all_models())

    def get_private_clouds(self):
        return self._total_count(models.PrivateCloud.get_all_models())

    def get_storages(self):
        return self._total_count(models.Storage.get_all_models())

    def get_users(self):
        return self.object.get_users().count()

    def _total_count(self, models):
        return sum(self._count_model(model) for model in models)

    def _count_model(self, model):
        qs = model.objects.filter(project=self.object).only('pk')
        qs = filter_queryset_for_user(qs, self.request.user)
        return qs.count()


class UserCountersView(BaseCounterView):
    """
    Count number of entities related to current user

    .. code-block:: javascript

        {
            "keys": 1,
            "hooks": 1
        }
    """

    def get_fields(self):
        return {
            'keys': self.get_keys,
            'hooks': self.get_hooks
        }

    def get_keys(self):
        return core_models.SshPublicKey.objects.filter(user_uuid=self.request.user.uuid.hex).count()

    def get_hooks(self):
        return core_managers.SummaryQuerySet(logging_models.BaseHook.get_all_models()).count()


class BaseServiceViewSet(core_mixins.EagerLoadMixin, core_views.ActionsViewSet):
    queryset = NotImplemented
    serializer_class = NotImplemented
    import_serializer_class = NotImplemented
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.BaseServiceFilter
    lookup_field = 'uuid'
    metadata_class = ActionsMetadata
    unsafe_methods_permissions = [permissions.is_owner, permissions.check_access_to_services_management]

    def list(self, request, *args, **kwargs):
        """
        To list all services without regard to its type, run **GET** against */api/services/* as an authenticated user.

        To list services of specific type issue **GET** to specific endpoint from a list above as a customer owner.
        Individual endpoint used for every service type.

        To create a service, issue a **POST** to specific endpoint from a list above as a customer owner.
        Individual endpoint used for every service type.

        You can create service based on shared service settings. Example:

        .. code-block:: http

            POST /api/digitalocean/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "Common DigitalOcean",
                "customer": "http://example.com/api/customers/1040561ca9e046d2b74268600c7e1105/",
                "settings": "http://example.com/api/service-settings/93ba615d6111466ebe3f792669059cb4/"
            }

        Or provide your own credentials. Example:

        .. code-block:: http

            POST /api/oracle/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "My Oracle",
                "customer": "http://example.com/api/customers/1040561ca9e046d2b74268600c7e1105/",
                "backend_url": "https://oracle.example.com:7802/em",
                "username": "admin",
                "password": "secret"
            }
        """
        return super(BaseServiceViewSet, self).list(request, *args, **kwargs)

    def _can_import(self):
        return self.import_serializer_class is not NotImplemented

    def get_serializer_class(self):
        serializer = super(BaseServiceViewSet, self).get_serializer_class()
        if self.action == 'link':
            serializer = self.import_serializer_class if self._can_import() else rf_serializers.Serializer

        return serializer

    def get_serializer_context(self):
        context = super(BaseServiceViewSet, self).get_serializer_context()
        # Viewset doesn't have object during schema generation
        if self.action == 'link' and self.lookup_field in self.kwargs:
            context['service'] = self.get_object()
        return context

    def get_import_context(self):
        return {}

    @detail_route()
    def managed_resources(self, request, uuid=None):
        service = self.get_object()
        backend = self.get_backend(service)

        try:
            resources = backend.get_managed_resources()
        except ServiceBackendNotImplemented:
            resources = []

        serializer = serializers.ManagedResourceSerializer(resources, many=True)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def _has_import_serializer_permission(request, view, obj=None):
        if not view._can_import():
            raise MethodNotAllowed(view.action)

    def _require_staff_for_shared_settings(request, view, obj=None):
        """ Allow to execute action only if service settings are not shared or user is staff """
        if obj is None:
            return

        if obj.settings.shared and not request.user.is_staff:
            raise PermissionDenied(_('Only staff users are allowed to import resources from shared services.'))

    @detail_route(methods=['get', 'post'])
    def link(self, request, uuid=None):
        """
        To get a list of resources available for import, run **GET** against */<service_endpoint>/link/*
        as an authenticated user.
        Optionally project_uuid parameter can be supplied for services requiring it like OpenStack.

        To import (link with Waldur) resource issue **POST** against the same endpoint with resource id.

        .. code-block:: http

            POST /api/openstack/08039f01c9794efc912f1689f4530cf0/link/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "backend_id": "bd5ec24d-9164-440b-a9f2-1b3c807c5df3",
                "project": "http://example.com/api/projects/e5f973af2eb14d2d8c38d62bcbaccb33/"
            }
        """

        service = self.get_object()

        if self.request.method == 'GET':
            try:
                backend = self.get_backend(service)
                try:
                    resources = backend.get_resources_for_import(**self.get_import_context())
                except ServiceBackendNotImplemented:
                    resources = []

                page = self.paginate_queryset(resources)
                if page is not None:
                    return self.get_paginated_response(page)

                return Response(resources)
            except (ServiceBackendError, ValidationError) as e:
                raise APIException(e)

        else:
            serializer = self.get_serializer(data=request.data)
            serializer.is_valid(raise_exception=True)

            try:
                resource = serializer.save()
            except ServiceBackendError as e:
                raise APIException(e)

            resource_imported.send(
                sender=resource.__class__,
                instance=resource,
            )

            return Response(serializer.data, status=status.HTTP_200_OK)

    link_permissions = [_has_import_serializer_permission, _require_staff_for_shared_settings]

    def get_backend(self, service):
        # project_uuid can be supplied in order to get a list of resources
        # available for import (link) based on project, depends on backend implementation
        project_uuid = self.request.query_params.get('project_uuid')
        if project_uuid:
            spl_class = SupportedServices.get_related_models(service)['service_project_link']
            try:
                spl = spl_class.objects.get(project__uuid=project_uuid, service=service)
            except spl_class.DoesNotExist:
                raise NotFound(_("Can't find project %s.") % project_uuid)
            else:
                return spl.get_backend()
        else:
            return service.get_backend()

    @detail_route(methods=['post'])
    def unlink(self, request, uuid=None):
        """
        Unlink all related resources, service project link and service itself.
        """
        service = self.get_object()
        service.unlink_descendants()
        self.perform_destroy(service)

        return Response(status=status.HTTP_204_NO_CONTENT)

    unlink_permissions = [
        _require_staff_for_shared_settings,
        permissions.check_access_to_services_management
    ]
    unlink.destructive = True


class BaseServiceProjectLinkViewSet(core_views.ActionsViewSet):
    queryset = NotImplemented
    serializer_class = NotImplemented
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.BaseServiceProjectLinkFilter
    unsafe_methods_permissions = [permissions.is_owner]
    disabled_actions = ['update', 'partial_update']

    def list(self, request, *args, **kwargs):
        """
        To get a list of connections between a project and an service, run **GET** against service_project_link_url
        as authenticated user. Note that a user can only see connections of a project where a user has a role.

        If service has `available_for_all` flag, project-service connections are created automatically.
        Otherwise, in order to be able to provision resources, service must first be linked to a project.
        To do that, **POST** a connection between project and a service to service_project_link_url
        as stuff user or customer owner.
        """
        return super(BaseServiceProjectLinkViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        To remove a link, issue **DELETE** to URL of the corresponding connection as stuff user or customer owner.
        """
        return super(BaseServiceProjectLinkViewSet, self).retrieve(request, *args, **kwargs)


class ResourceViewMetaclass(type):
    """ Store view in registry """
    def __new__(cls, name, bases, args):
        resource_view = super(ResourceViewMetaclass, cls).__new__(cls, name, bases, args)
        queryset = args.get('queryset')
        if hasattr(queryset, 'model') and not issubclass(queryset.model, models.SubResource):
            SupportedServices.register_resource_view(queryset.model, resource_view)
        return resource_view


class BaseServicePropertyViewSet(viewsets.ReadOnlyModelViewSet):
    filter_class = filters.BaseServicePropertyFilter


class AggregatedStatsView(views.APIView):
    """
    Quotas and quotas usage aggregated by projects/customers.

    Available request parameters:
        - ?aggregate=aggregate_model_name (default: 'customer'.
          Have to be from list: 'customer', 'project')
        - ?uuid=uuid_of_aggregate_model_object (not required. If this parameter will be defined -
          result will contain only object with given uuid)
        - ?quota_name - optional list of quota names, for example ram, vcpu, storage
    """

    def get(self, request, format=None):
        serializer = serializers.AggregateSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)

        quota_names = request.query_params.getlist('quota_name')
        if len(quota_names) == 0:
            quota_names = None
        querysets = serializer.get_service_project_links(request.user)

        total_sum = QuotaModelMixin.get_sum_of_quotas_for_querysets(querysets, quota_names)
        total_sum = sort_dict(total_sum)
        return Response(total_sum, status=status.HTTP_200_OK)


class QuotaTimelineStatsView(views.APIView):
    """
    Historical data of quotas and quotas usage aggregated by projects/customers.

    Available request parameters:

    - ?from=timestamp (default: now - 1 day, for example: 1415910025)
    - ?to=timestamp (default: now, for example: 1415912625)
    - ?interval (default: day. Has to be from list: hour, day, week, month)
    - ?item=<quota_name>. If this parameter is not defined - endpoint will return data for all items.
    - ?aggregate=aggregate_model_name (default: 'customer'. Have to be from list: 'customer', 'project')
    - ?uuid=uuid_of_aggregate_model_object (not required. If this parameter is defined, result will contain only object with given uuid)

    Answer will be list of dictionaries with fields, determining time frame. It's size is equal to interval parameter.
    Values within each bucket are averaged for each project and then all projects metrics are summarized.

    Value fields include:

    - vcpu_limit - virtual CPUs quota
    - vcpu_usage - virtual CPUs usage
    - ram_limit - RAM quota limit, in MiB
    - ram_usage - RAM usage, in MiB
    - storage_limit - volume storage quota limit, in MiB
    - storage_usage - volume storage quota consumption, in MiB
    """

    def get(self, request, format=None):
        scopes = self.get_quota_scopes(request)
        ranges = self.get_ranges(request)
        items = request.query_params.getlist('item') or self.get_all_spls_quotas()

        collector = QuotaTimelineCollector()
        for item in items:
            for scope in scopes:
                values = self.get_stats_for_scope(item, scope, ranges)
                for (end, start), (limit, usage) in zip(ranges, values):
                    collector.add_quota(start, end, item, limit, usage)

        stats = list(map(sort_dict, collector.to_dict()))[::-1]
        return Response(stats, status=status.HTTP_200_OK)

    def get_quota_scopes(self, request):
        serializer = serializers.AggregateSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        scopes = sum([list(qs) for qs in serializer.get_service_project_links(request.user)], [])
        # XXX: quick and dirty hack for OpenStack: use tenants instead of SPLs as quotas scope.
        new_scopes = []
        for index, scope in enumerate(scopes):
            if scope.service.settings.type == 'OpenStack':
                new_scopes += list(scope.tenants.all())
            else:
                new_scopes.append(scope)
        return new_scopes

    def get_all_spls_quotas(self):
        # XXX: quick and dirty hack for OpenStack: use tenants instead of SPLs as quotas scope.
        spl_models = [m if m.__name__ != 'OpenStackServiceProjectLink' else m.tenants.field.model
                      for m in models.ServiceProjectLink.get_all_models()]
        return sum([spl_model.get_quotas_names() for spl_model in spl_models], [])

    def get_stats_for_scope(self, quota_name, scope, dates):
        stats_data = []
        try:
            quota = scope.quotas.get(name=quota_name)
        except Quota.DoesNotExist:
            return stats_data
        versions = Version.objects.get_for_object(quota).select_related('revision').filter(
            revision__date_created__lte=dates[0][0]).iterator()
        version = None
        for end, start in dates:
            try:
                while version is None or version.revision.date_created > end:
                    version = next(versions)
                stats_data.append((version._object_version.object.limit,
                                   version._object_version.object.usage))
            except StopIteration:
                break

        return stats_data

    def get_ranges(self, request):
        mapped = {
            'start_time': request.query_params.get('from'),
            'end_time': request.query_params.get('to'),
            'interval': request.query_params.get('interval')
        }
        data = {key: val for (key, val) in mapped.items() if val}

        serializer = core_serializers.TimelineSerializer(data=data)
        serializer.is_valid(raise_exception=True)

        date_points = serializer.get_date_points()
        reversed_dates = date_points[::-1]
        ranges = list(zip(reversed_dates[:-1], reversed_dates[1:]))
        return ranges


class QuotaTimelineCollector(object):
    """
    Helper class for QuotaTimelineStatsView.
    Aggregate quotas grouped by date range and quota name.
    Example output rendering:

    .. code-block:: javascript

        [
            {
                "from": start,
                "to" end,
                "vcpu_limit": 10,
                "vcpu_usage": 5,
                "ram_limit": 4000,
                "ran_usage": 1000
            }
        ]
    """

    def __init__(self):
        self.ranges = set()
        self.items = set()
        self.limits = defaultdict(int)
        self.usages = defaultdict(int)

    def add_quota(self, start, end, item, limit, usage):
        key = (start, end, item)
        if limit == -1 or self.limits[key] == -1:
            self.limits[key] = -1
        else:
            self.limits[key] += limit
        self.usages[key] += usage
        self.ranges.add((start, end))
        self.items.add(item)

    def to_dict(self):
        table = []
        for start, end in sorted(self.ranges):
            row = {
                'from': datetime_to_timestamp(start),
                'to': datetime_to_timestamp(end)
            }
            for item in sorted(self.items):
                key = (start, end, item)
                row['%s_limit' % item] = self.limits[key]
                row['%s_usage' % item] = self.usages[key]
            table.append(row)
        return table


def check_resource_backend_id(resource):
    if not resource.backend_id:
        raise ValidationError(_('Resource does not have backend ID.'))


class ResourceViewSet(core_mixins.ExecutorMixin, core_views.ActionsViewSet):
    """ Basic view set for all resource view sets. """
    lookup_field = 'uuid'
    filter_backends = (filters.GenericRoleFilter, DjangoFilterBackend)
    metadata_class = ActionsMetadata
    unsafe_methods_permissions = [permissions.is_administrator]
    update_validators = partial_update_validators = [core_validators.StateValidator(models.NewResource.States.OK)]
    destroy_validators = [core_validators.StateValidator(models.NewResource.States.OK, models.NewResource.States.ERRED)]

    @detail_route(methods=['post'])
    def pull(self, request, uuid=None):
        if not self.pull_executor:
            return Response({'detail': _('Pull operation is not implemented.')},
                            status=status.HTTP_409_CONFLICT)
        self.pull_executor.execute(self.get_object())
        return Response({'detail': _('Pull operation was successfully scheduled.')}, status=status.HTTP_202_ACCEPTED)

    pull_executor = NotImplemented
    pull_validators = [
        core_validators.StateValidator(models.NewResource.States.OK, models.NewResource.States.ERRED),
        check_resource_backend_id,
    ]


class BaseResourceViewSet(six.with_metaclass(ResourceViewMetaclass, ResourceViewSet)):
    pass


class ImportableResourceViewSet(BaseResourceViewSet):
    """
    This viewset enables uniform implementation of resource import.

    Comparing to the previous approach when import endpoint was defined in service viewset,
    this class assumes that resource-specific import options are defined in resource viewset.

    Consider the following example:

    importable_resources_backend_method = 'get_tenants_for_import'
    importable_resources_serializer_class = serializers.TenantImportableSerializer
    importable_resources_permissions = [structure_permissions.is_staff]
    import_resource_serializer_class = serializers.TenantImportSerializer
    import_resource_permissions = [structure_permissions.is_staff]
    import_resource_executor = executors.TenantImportExecutor

    Note that there are only 3 mandatory parameters:
    * importable_resources_backend_method
    * importable_resources_serializer_class
    * import_resource_serializer_class
    """
    import_resource_executor = None

    @list_route(methods=['get'])
    def importable_resources(self, request):
        serializer = self.get_serializer(data=request.GET)
        serializer.is_valid(raise_exception=True)
        service_project_link = serializer.validated_data['service_project_link']

        backend = service_project_link.get_backend()
        resources = getattr(backend, self.importable_resources_backend_method)()
        serializer = self.get_serializer(resources, many=True)
        page = self.paginate_queryset(serializer.data)
        if page is not None:
            return self.get_paginated_response(page)

        return Response(data=serializer.data, status=status.HTTP_200_OK)

    @list_route(methods=['post'])
    def import_resource(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            resource = serializer.save()
        except IntegrityError:
            raise rf_serializers.ValidationError(_('Resource is already registered.'))
        else:
            resource_imported.send(
                sender=resource.__class__,
                instance=resource,
            )
        if self.import_resource_executor:
            self.import_resource_executor.execute(resource)

        return Response(data=serializer.data, status=status.HTTP_201_CREATED)


class ServiceCertificationViewSet(core_views.ActionsViewSet):
    lookup_field = 'uuid'
    metadata_class = ActionsMetadata
    unsafe_methods_permissions = [permissions.is_staff]
    serializer_class = serializers.ServiceCertificationSerializer
    queryset = models.ServiceCertification.objects.all()


class DivisionViewSet(core_views.ReadOnlyActionsViewSet):
    queryset = models.Division.objects.all().order_by('name')
    serializer_class = serializers.DivisionSerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.DivisionFilter



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-13 11:17
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0055_offeringfile'),
    ]

    operations = [
        migrations.AddField(
            model_name='plan',
            name='max_amount',
            field=models.PositiveSmallIntegerField(blank=True, help_text='Maximum number of plans that could be active. Plan is disabled when maximum amount is reached.', null=True, validators=[django.core.validators.MinValueValidator(1)]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.contrib.contenttypes.models import ContentType


def initialize_subnet_count_quota(apps, schema_editor):
    Tenant = apps.get_model('openstack', 'Tenant')
    Quota = apps.get_model('quotas', 'Quota')
    tenant_content_type = ContentType.objects.get_for_model(Tenant)
    quota_name = 'subnet_count'
    tenant_ids = Quota.objects.filter(name=quota_name, content_type_id=tenant_content_type.id). \
        values_list('object_id', flat=True)

    for tenant in Tenant.objects.exclude(id__in=tenant_ids).iterator():
        Quota.objects.create(
            name=quota_name,
            limit=10,
            usage=tenant.networks.count(),
            content_type_id=tenant_content_type.id,
            object_id=tenant.id)


def initialize_network_count_quota(apps, schema_editor):
    Tenant = apps.get_model('openstack', 'Tenant')
    SubNet = apps.get_model('openstack', 'SubNet')
    Quota = apps.get_model('quotas', 'Quota')
    tenant_content_type = ContentType.objects.get_for_model(Tenant)
    quota_name = 'network_count'
    tenant_ids = Quota.objects.filter(name=quota_name, content_type_id=tenant_content_type.id).\
        values_list('object_id', flat=True)

    for tenant in Tenant.objects.exclude(id__in=tenant_ids).iterator():
        subnet_count = SubNet.objects.filter(network__pk__in=tenant.networks.values('pk')).count()
        Quota.objects.create(
            name=quota_name,
            limit=10,
            usage=subnet_count,
            content_type_id=tenant_content_type.id,
            object_id=tenant.id)


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('quotas', '0004_quota_threshold'),
        ('openstack', '0028_security_group_as_a_resource'),
    ]

    operations = [
        migrations.RunPython(initialize_network_count_quota),
        migrations.RunPython(initialize_subnet_count_quota),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-13 16:19
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0033_plancomponent_billing_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='ComponentQuota',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('limit', models.PositiveIntegerField(default=-1)),
                ('usage', models.PositiveIntegerField(default=0)),
                ('component', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.PlanComponent')),
            ],
        ),
        migrations.AlterModelOptions(
            name='order',
            options={'ordering': ('created',), 'verbose_name': 'Order'},
        ),
        migrations.AlterModelOptions(
            name='orderitem',
            options={'ordering': ('created',), 'verbose_name': 'Order item'},
        ),
        migrations.AddField(
            model_name='componentquota',
            name='order_item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='quotas', to='marketplace.OrderItem'),
        ),
        migrations.AlterUniqueTogether(
            name='componentquota',
            unique_together=set([('order_item', 'component')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-03 06:29
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0080_resource_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='category',
            name='icon',
            field=models.FileField(blank=True, null=True, upload_to='marketplace_category_icons', validators=[waldur_core.core.validators.FileTypeValidator(allowed_types=['image/png', 'image/jpeg', 'image/svg', 'image/svg+xml', 'image/x-icon'])]),
        ),
        migrations.AlterField(
            model_name='offering',
            name='thumbnail',
            field=models.FileField(blank=True, null=True, upload_to='marketplace_service_offering_thumbnails', validators=[waldur_core.core.validators.FileTypeValidator(allowed_types=['image/png', 'image/jpeg', 'image/svg', 'image/svg+xml', 'image/x-icon'])]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-31 08:19
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0029_issue_resolution_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='issues', to='support.Template'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-23 11:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0002_immutable_default_json'),
        ('marketplace', '0024_rename_screenshot'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='allowed_customers',
            field=models.ManyToManyField(blank=True, to='structure.Customer'),
        ),
        migrations.AddField(
            model_name='offering',
            name='shared',
            field=models.BooleanField(default=False, help_text='Accessible to all customers.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-05 13:07
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('support_invoices', '0001_initial'),
    ]

    operations = [
        migrations.DeleteModel(
            name='RequestBasedOffering',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0029_tenant_quotas'),
    ]

    operations = [
        migrations.AddField(
            model_name='subnet',
            name='dns_nameservers',
            field=waldur_core.core.fields.JSONField(default=[], help_text='List of DNS name servers associated with the subnet.'),
        ),
    ]



from celery import chain

from waldur_core.core import executors
from waldur_core.core.tasks import StateTransitionTask, BackendMethodTask
from waldur_core.core.models import RuntimeStateMixin
from waldur_core.core import utils as core_utils
from waldur_core.structure import executors as structure_executors

from . import tasks, models


class DropletCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        return chain(
            BackendMethodTask().si(
                serialized_droplet, 'create_droplet',
                state_transition='begin_creating',
                runtime_state='provisioning',
                success_runtime_state=RuntimeStateMixin.RuntimeStates.ONLINE,
                **kwargs),
            tasks.WaitForActionComplete().s(serialized_droplet).set(countdown=30))


class DropletDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        if droplet.backend_id:
            return BackendMethodTask().si(
                serialized_droplet, 'destroy', state_transition='begin_deleting')
        else:
            return StateTransitionTask().si(serialized_droplet, state_transition='begin_deleting')


class DropletStopExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        return chain(
            BackendMethodTask().si(
                serialized_droplet, 'stop',
                state_transition='begin_updating',
                success_runtime_state=models.Droplet.RuntimeStates.OFFLINE),
            tasks.WaitForActionComplete().s(serialized_droplet).set(countdown=10))


class DropletStartExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        return chain(
            BackendMethodTask().si(
                serialized_droplet, 'start',
                state_transition='begin_updating',
                success_runtime_state=models.Droplet.RuntimeStates.ONLINE),
            tasks.WaitForActionComplete().s(serialized_droplet).set(countdown=10))


class DropletRestartExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        return chain(
            BackendMethodTask().si(
                serialized_droplet, 'restart',
                state_transition='begin_updating',
                success_runtime_state=models.Droplet.RuntimeStates.ONLINE),
            tasks.WaitForActionComplete().s(serialized_droplet).set(countdown=10))


class DropletResizeExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, droplet, serialized_droplet, **kwargs):
        size = kwargs.pop('size')
        disk = kwargs.pop('disk')
        return chain(
            BackendMethodTask().si(
                serialized_droplet, 'resize',
                state_transition='begin_updating',
                runtime_state='resizing',
                success_runtime_state=models.Droplet.RuntimeStates.ONLINE,
                backend_size_id=size.backend_id,
                disk=disk),
            tasks.WaitForActionComplete().s(serialized_droplet).set(countdown=10),
            tasks.LogDropletResized().si(serialized_droplet, core_utils.serialize_instance(size)))


class DigitalOceanCleanupExecutor(structure_executors.BaseCleanupExecutor):
    executors = (
        (models.Droplet, DropletDeleteExecutor),
    )



from __future__ import unicode_literals

from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets, permissions, response, status
from rest_framework.decorators import list_route

from waldur_core.core import views as core_views
from waldur_core.structure import filters as structure_filters, permissions as structure_permissions

from .log import event_logger
from . import filters, models, serializers, executors


class PackageTemplateViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.PackageTemplate.objects.all().order_by('name')
    serializer_class = serializers.PackageTemplateSerializer
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.PackageTemplateFilter


class OpenStackPackageViewSet(core_views.ActionsViewSet):
    queryset = models.OpenStackPackage.objects.all()
    serializer_class = serializers.OpenStackPackageSerializer
    lookup_field = 'uuid'
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.OpenStackPackageFilter
    disabled_actions = ['update', 'partial_update', 'destroy']

    def create(self, request, *args, **kwargs):
        # package creation is a little bit tricky:
        # We need to use OpenStackPackageCreateSerializer to create package
        # And OpenStackPackageSerializer to display it.
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        package = serializer.save()
        skip = (settings.WALDUR_CORE['ONLY_STAFF_MANAGES_SERVICES'] and
                serializer.validated_data['skip_connection_extnet'])
        executors.OpenStackPackageCreateExecutor.execute(package, skip_connection_extnet=skip)

        display_serializer = serializers.OpenStackPackageSerializer(instance=package, context={'request': request})
        return response.Response(display_serializer.data, status=status.HTTP_201_CREATED)

    create_serializer_class = serializers.OpenStackPackageCreateSerializer
    create_permissions = [structure_permissions.check_access_to_services_management]

    @list_route(methods=['post'])
    def change(self, request, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        package = serializer.validated_data['package']
        new_template = serializer.validated_data['template']
        service_settings = package.service_settings
        tenant = package.tenant

        event_logger.openstack_package.info(
            'Tenant package change has been scheduled. '
            'Old value: %s, new value: {package_template_name}' % package.template.name,
            event_type='openstack_package_change_scheduled',
            event_context={
                'tenant': tenant,
                'package_template_name': new_template.name,
                'service_settings': service_settings,
            })

        executors.OpenStackPackageChangeExecutor.execute(tenant,
                                                         new_template=new_template,
                                                         old_package=package,
                                                         service_settings=service_settings)

        return response.Response({'detail': _('OpenStack package extend has been scheduled')},
                                 status=status.HTTP_202_ACCEPTED)

    change_serializer_class = serializers.OpenStackPackageChangeSerializer
    change_permissions = create_permissions

    @list_route(methods=['post'])
    def assign(self, request, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response({'detail': _('OpenStack package has been assigned')}, status=status.HTTP_200_OK)

    assign_serializer_class = serializers.OpenStackPackageAssignSerializer
    assign_permissions = [structure_permissions.is_staff]



import logging

from waldur_azure.client import AzureClient, AzureBackendError, reraise
from waldur_core.structure import ServiceBackend

from . import models

logger = logging.getLogger(__name__)


class AzureBackend(ServiceBackend):
    def __init__(self, settings):
        self.settings = settings
        self.client = AzureClient(settings)

    def ping(self, raise_exception=False):
        try:
            self.client.list_locations()
        except AzureBackendError as e:
            if raise_exception:
                reraise(e)
            return False
        else:
            return True

    def pull_service_properties(self):
        self.pull_locations()

        location = models.Location.objects.filter(settings=self.settings).last()
        self.pull_sizes(location)

    def pull_locations(self):
        cached_locations = {
            location.backend_id: location
            for location in models.Location.objects.filter(settings=self.settings)
        }

        backend_locations = {
            location.name: location
            for location in self.client.list_locations()
        }

        resource_group_locations = self.client.get_resource_group_locations()

        new_locations = {
            location for name, location in backend_locations.items()
            if name not in cached_locations
        }

        stale_locations = {
            location for name, location in cached_locations.items()
            if name not in backend_locations
        }

        for backend_location in new_locations:
            models.Location.objects.create(
                backend_id=backend_location.name,
                name=backend_location.display_name,
                latitude=backend_location.latitude,
                longitude=backend_location.longitude,
                settings=self.settings,
            )

        for cached_location in stale_locations:
            cached_location.delete()

        models.Location.objects.filter(name__in=resource_group_locations).update(enabled=True)
        models.Location.objects.exclude(name__in=resource_group_locations).update(enabled=False)

    def pull_public_ips(self, service_project_link):
        locations = {
            location.backend_id: location
            for location in models.Location.objects.filter(settings=self.settings)
        }

        cached_public_ips = {
            public_ip.backend_id: public_ip
            for public_ip in models.PublicIP.objects.filter(service_project_link=service_project_link)
        }

        backend_public_ips = {
            public_ip.name: public_ip
            for public_ip in self.client.list_all_public_ips()
        }

        new_public_ips = {
            public_ip for name, public_ip in backend_public_ips.items()
            if name not in cached_public_ips
        }

        stale_public_ips = {
            public_ip for name, public_ip in cached_public_ips.items()
            if name not in backend_public_ips
        }

        for backend_public_ip in new_public_ips:
            models.PublicIP.objects.create(
                backend_id=backend_public_ip.name,
                name=backend_public_ip.name,
                service_project_link=service_project_link,
                location=locations.get(backend_public_ip.location),
                state=models.PublicIP.States.OK,
            )

        for cached_public_ip in stale_public_ips:
            cached_public_ip.delete()

    def pull_sizes(self, location):
        cached_sizes = {
            size.backend_id: size
            for size in models.Size.objects.filter(settings=self.settings)
        }

        backend_sizes = {
            size.name: size
            for size in self.client.list_virtual_machine_sizes(location.backend_id)
        }

        new_sizes = {
            size for name, size in backend_sizes.items()
            if name not in cached_sizes
        }

        stale_sizes = {
            size for name, size in cached_sizes.items()
            if name not in backend_sizes
        }

        for backend_size in new_sizes:
            models.Size.objects.create(
                backend_id=backend_size.name,
                settings=self.settings,
                **backend_size.as_dict()
            )

        for cached_size in stale_sizes:
            cached_size.delete()

    def pull_resource_groups(self, service_project_link):
        cached_groups = {
            group.backend_id: group
            for group in models.ResourceGroup.objects.filter(service_project_link=service_project_link)
        }

        backend_groups = {
            group.name: group
            for group in self.client.list_resource_groups()
        }

        new_groups = {
            group for name, group in backend_groups.items()
            if name not in cached_groups
        }

        stale_groups = {
            group for name, group in cached_groups.items()
            if name not in backend_groups
        }

        locations = {
            location.backend_id: location
            for location in models.Location.objects.filter(settings=self.settings)
        }

        for backend_group in new_groups:
            models.ResourceGroup.objects.create(
                backend_id=backend_group.id,
                name=backend_group.name,
                service_project_link=service_project_link,
                location=locations.get(backend_group.location),
                state=models.ResourceGroup.States.OK,
            )

        for cached_group in stale_groups:
            cached_group.delete()

    def create_resource_group(self, resource_group):
        backend_resource_group = self.client.create_resource_group(
            location=resource_group.location.backend_id,
            resource_group_name=resource_group.name
        )
        resource_group.backend_id = backend_resource_group.id
        resource_group.save()

    def create_storage_account(self, storage_account):
        poller = self.client.create_storage_account(
            location=storage_account.resource_group.location.backend_id,
            resource_group_name=storage_account.resource_group.name,
            account_name=storage_account.name,
        )
        backend_storage_account = poller.result()
        storage_account.backend_id = backend_storage_account.id
        storage_account.save()

    def delete_resource_group(self, resource_group):
        self.client.delete_resource_group(resource_group.backend_id)

    def create_network(self, network):
        poller = self.client.create_network(
            location=network.resource_group.location.backend_id,
            resource_group_name=network.resource_group.name,
            network_name=network.name,
            cidr=network.cidr,
        )
        backend_network = poller.result()
        network.backend_id = backend_network.id
        network.save()

    def create_subnet(self, subnet):
        poller = self.client.create_subnet(
            resource_group_name=subnet.resource_group.name,
            network_name=subnet.network.name,
            subnet_name=subnet.name,
            cidr=subnet.cidr,
        )
        backend_subnet = poller.result()
        subnet.backend_id = backend_subnet.id
        subnet.save()

    def pull_network_interface(self, nic):
        backend_nic = self.client.get_network_interface(
            resource_group_name=nic.resource_group.name,
            network_interface_name=nic.name,
        )
        nic.ip_address = backend_nic.ip_configurations[0].private_ip_address
        nic.save()

    def create_network_interface(self, nic):
        poller = self.client.create_network_interface(
            location=nic.resource_group.location.backend_id,
            resource_group_name=nic.resource_group.name,
            interface_name=nic.name,
            subnet_id=nic.subnet.backend_id,
            config_name=nic.config_name,
            public_ip_id=nic.public_ip and nic.public_ip.backend_id,
        )
        backend_nic = poller.result()
        nic.backend_id = backend_nic.id
        nic.save()

    def create_virtual_machine(self, vm):
        poller = self.client.create_virtual_machine(
            location=vm.resource_group.location.backend_id,
            resource_group_name=vm.resource_group.name,
            vm_name=vm.name,
            size_name=vm.size.name,
            nic_id=vm.network_interface.backend_id,
            image_reference={
                'sku': vm.image.sku,
                'publisher': vm.image.publisher,
                'version': vm.image.version,
                'offer': vm.image.name,
            },
            username=vm.username,
            password=vm.password,
            ssh_key=vm.ssh_key and vm.ssh_key.public_key or None,
            custom_data=vm.user_data,
        )
        backend_vm = poller.result()
        vm.backend_id = backend_vm.id
        vm.runtime_state = backend_vm.provisioning_state
        vm.save()

    def start_virtual_machine(self, virtual_machine):
        poller = self.client.start_virtual_machine(
            resource_group_name=virtual_machine.resource_group.name,
            vm_name=virtual_machine.name,
        )
        poller.wait()

    def restart_virtual_machine(self, virtual_machine):
        poller = self.client.restart_virtual_machine(
            resource_group_name=virtual_machine.resource_group.name,
            vm_name=virtual_machine.name,
        )
        poller.wait()

    def stop_virtual_machine(self, virtual_machine):
        poller = self.client.stop_virtual_machine(
            resource_group_name=virtual_machine.resource_group.name,
            vm_name=virtual_machine.name,
        )
        poller.wait()

    def delete_virtual_machine(self, virtual_machine):
        poller = self.client.delete_virtual_machine(
            resource_group_name=virtual_machine.resource_group.name,
            vm_name=virtual_machine.name,
        )
        poller.wait()

    def create_ssh_security_group(self, network_security_group):
        poller = self.client.create_ssh_security_group(
            location=network_security_group.resource_group.location.backend_id,
            resource_group_name=network_security_group.resource_group.name,
            network_security_group_name=network_security_group.name,
        )
        backend_group = poller.result()
        network_security_group.backend_id = backend_group.id
        network_security_group.save()

    def create_public_ip(self, public_ip):
        poller = self.client.create_public_ip(
            location=public_ip.resource_group.location.backend_id,
            resource_group_name=public_ip.resource_group.name,
            public_ip_address_name=public_ip.name,
        )
        backend_public_ip = poller.result()
        public_ip.backend_id = backend_public_ip.id
        public_ip.runtime_state = backend_public_ip.provisioning_state
        public_ip.save()

    def delete_public_ip(self, public_ip):
        poller = self.client.delete_public_ip(
            resource_group_name=public_ip.resource_group.name,
            public_ip_address_name=public_ip.name,
        )
        poller.wait()

    def pull_public_ip_address(self, public_ip):
        backend_public_ip = self.client.get_public_ip(
            resource_group_name=public_ip.resource_group.name,
            public_ip_address_name=public_ip.name,
        )
        public_ip.ip_address = backend_public_ip.ip_address
        public_ip.runtime_state = backend_public_ip.provisioning_state
        public_ip.save()

    def create_pgsql_server(self, server):
        backend_server = self.client.create_sql_server(
            location=server.resource_group.location.backend_id,
            resource_group_name=server.resource_group.name,
            server_name=server.name,
            username=server.username,
            password=server.password,
            storage_mb=server.storage_mb,
            sku={
                'name': 'B_Gen5_1',
                'tier': 'Basic',
                'family': 'Gen5',
                'capacity': 1,
            },
        )
        server.backend_id = backend_server.id
        server.fqdn = backend_server.fully_qualified_domain_name
        server.save()

        self.client.create_sql_firewall_rule(
            resource_group_name=server.resource_group.name,
            server_name=server.name,
            firewall_rule_name='firewall{}'.format(server.name),
            start_ip_address='0.0.0.0',  # nosec
            end_ip_address='255.255.255.255',  # nosec
        )

    def delete_pgsql_server(self, server):
        poller = self.client.delete_sql_server(
            resource_group_name=server.resource_group.name,
            server_name=server.name,
        )
        poller.wait()

    def create_pgsql_database(self, database):
        poller = self.client.create_sql_database(
            resource_group_name=database.server.resource_group.name,
            server_name=database.server.name,
            database_name=database.name,
            charset=database.charset,
            collation=database.collation,
        )
        backend_database = poller.result()
        database.backend_id = backend_database.id
        database.save()

    def delete_pgsql_database(self, database):
        poller = self.client.delete_sql_database(
            resource_group_name=database.server.resource_group.name,
            server_name=database.server.name,
            database_name=database.name,
        )
        poller.wait()



import logging

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

from waldur_core.core import utils as core_utils
from waldur_mastermind.common import mixins as common_mixins
from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace.models import Resource
from waldur_mastermind.support import models as support_models


logger = logging.getLogger(__name__)


def is_request_based(offering):
    return Resource.objects.filter(scope=offering).exists()


def get_offering_items():
    model_type = ContentType.objects.get_for_model(support_models.Offering)
    return invoice_models.GenericInvoiceItem.objects.filter(content_type=model_type)


def component_usage_register(component_usage):
    from waldur_mastermind.support_invoices.registrators import OfferingRegistrator

    plan_period = component_usage.plan_period
    if not plan_period:
        logger.warning('Skipping processing of component usage with ID %s because '
                       'plan period is not defined.', component_usage.id)
        return
    plan = plan_period.plan

    try:
        plan_component = plan.components.get(component=component_usage.component)
        item = invoice_models.GenericInvoiceItem.objects.get(scope=component_usage.resource.scope,
                                                             details__plan_period_id=plan_period.id,
                                                             details__plan_component_id=plan_component.id)
        item.quantity = component_usage.usage
        item.unit_price = plan_component.price
        item.save()
    except invoice_models.GenericInvoiceItem.DoesNotExist:
        offering = component_usage.resource.scope
        customer = offering.project.customer
        invoice, created = registrators.RegistrationManager.get_or_create_invoice(customer, component_usage.date)

        details = OfferingRegistrator().get_component_details(offering, plan_component)
        details['plan_period_id'] = plan_period.id
        offering_component = plan_component.component

        month_start = core_utils.month_start(component_usage.date)
        month_end = core_utils.month_end(component_usage.date)

        start = month_start if not component_usage.plan_period.start else \
            max(component_usage.plan_period.start, month_start)
        end = month_end if not component_usage.plan_period.end else \
            min(component_usage.plan_period.end, month_end)

        invoice_models.GenericInvoiceItem.objects.create(
            content_type=ContentType.objects.get_for_model(offering),
            object_id=offering.id,
            project=offering.project,
            invoice=invoice,
            start=start,
            end=end,
            details=details,
            unit_price=plan_component.price,
            quantity=component_usage.usage,
            unit=common_mixins.UnitPriceMixin.Units.QUANTITY,
            product_code=offering_component.product_code or plan.product_code,
            article_code=offering_component.article_code or plan.article_code,
        )

    except marketplace_models.PlanComponent.DoesNotExist:
        logger.warning('Plan component for usage component %s is not found.', component_usage.id)
    except invoice_models.GenericInvoiceItem.MultipleObjectsReturned:
        logger.warning('Skipping the invoice item unit price update '
                       'because multiple GenericInvoiceItem objects found. Scope: %s %s, date: %s.',
                       component_usage.resource.content_type,
                       component_usage.resource.object_id,
                       component_usage.date)


def resource_created_or_switched_plan(resource, date):
    """Check whether Resource has been created or updated after a given date.

        :param resource: marketplace resource.
        :param date: date.
        :return: tuple (has_been_created, has_been_switched).
    """

    create = False
    switch = False

    try:
        marketplace_models.OrderItem.objects.get(resource=resource,
                                                 type=marketplace_models.OrderItem.Types.CREATE,
                                                 activated__gte=date)
        create = True
    except ObjectDoesNotExist:
        pass
    except MultipleObjectsReturned:
        logger.debug('Multiple order items are found. resource ID: %s, type CREATE, activated__gte=%s',
                     resource.id, date)

    try:
        marketplace_models.OrderItem.objects.get(resource=resource,
                                                 type=marketplace_models.OrderItem.Types.UPDATE,
                                                 activated__gte=date)
        switch = True
    except ObjectDoesNotExist:
        pass
    except MultipleObjectsReturned:
        switch = True

    return create, switch



import collections
import logging

from django.conf import settings
import requests

UserDetails = collections.namedtuple('UserDetails', ('full_name', 'native_name', 'organization', 'job_title'))


class BCCException(Exception):
    default_code = 400
    default_detail = 'BCC authentication client error'

    def __init__(self, detail=None, code=None):
        self.detail = detail or self.default_detail
        self.code = code or self.default_code

    def __str__(self):
        return self.detail


def get_user_details(nid, vno):
    """
    Fetch user details from PayFixation API using configured credentials.
    :param nid: NID
    :type nid: string
    :param vno: Voucher No
    :type vno: string
    :return: user details
    :rtype: UserDetails
    """
    conf = settings.WALDUR_AUTH_BCC
    url = conf['BASE_API_URL']
    params = {
        'username': conf['USERNAME'],
        'password': conf['PASSWORD'],
        'nid': nid,
        'vno': vno,
        'type': 'bcc',
    }

    try:
        response = requests.get(url, params=params)
    except requests.RequestException as e:
        logging.warning('Unable to get user details from PayFixation API. Exception: %s', e)
        raise BCCException(detail=e.message)

    if response.status_code != 200:
        raise BCCException(code=response.status_code)

    try:
        data = response.json()
    except ValueError:
        raise BCCException('Unable to parse JSON.')

    error = data.get('error')
    if error:
        raise BCCException(detail=error)

    name = data['nameen']
    if not name:
        raise BCCException(detail='Invalid input parameters.')

    return UserDetails(
        full_name=data['nameen'],
        native_name=data['namebn'],
        job_title=data['desig'],
        organization=data['office'],
    )
